import logging
import traceback

import requests
from ..config import Config
from .base import CallerBase
from ..entity.answer import Answer
from ..entity.exception import BfEngineException, ErrorCode
from ..logger import log as logging


class KnowledgeGraphCaller(CallerBase):
    """
    知识图谱
    """

    def __init__(self, app_id):
        super().__init__(app_id)
        # 导入数据
        self.app_id = app_id
        base_url = Config.base_url
        self.data_load_url = '{}/xeonKgDal/{}/importJson'.format(base_url, app_id)
        # 触发异步训练
        self.trigger_train_url = '{}/xeonKgDal/{}/ftTrainTrigger?env=sandbox'.format(base_url, app_id)
        # 通知出话层更新数据
        self.sync_data_url = '{}/api/v1/kbqa/syncData?appid={}'.format(base_url, app_id)
        # 线下数据同步到线上
        self.sync_sandbox_url = '{}/xeonKgDal/{}/synchronizeSandbox'.format(base_url, app_id)
        # 预测出话接口
        self.predict_url = '{}/api/v1/kbqa'.format(base_url)
        # 获取训练结果
        self.train_status_url = '{}/xeonKgDal/syncHandler/{}/getSyncRecord'.format(base_url, app_id)

        self.file_load_url = '{}/xeonKgDal/{}/import/all'.format(base_url, app_id)
        self.new_file_load_url = '{}/xeonKgDal/{}/import/all/v1'.format(base_url, app_id)
        # 检查实体是否重复
        self.check_entity_repeat_url \
            = '{}/xeonKgDal/entityManager/{}/checkRepeat/entity?entityName={}&entityId=&entityType=1'
        # 获取实体 app_id, parent_entity_id, keyword
        self.load_entity_url \
            = '{}/xeonKgDal/entityProfile/{}/showEntityProfileDetail/{}?page=1&entityListKeyword={}'
        # 新增一级实体
        self.add_entity_url = '{}/xeonKgDal/entityManager/{}/addEntity/root'.format(base_url, app_id)
        # 获取属性
        self.load_property_url = '{}/xeonKgDal/{}/property/allCommon?page=1&keyword={}'
        # 新增屬性
        self.add_property_url = '{}/xeonKgDal/{}/property'.format(base_url, app_id)
        # 查询属性值
        self.load_relation_url = '{}/xeonKgDal/relation/{}/getDataOverViewRelationByRootEntity/{}?page=1&limit=20&entityListKeyword='
        # 新增属性值
        self.add_relation_url = '{}/xeonKgDal/relation/{}/addRelation'.format(base_url, app_id)
        # 更新属性值
        self.update_relation_url = '{}/xeonKgDal/relation/{}/updateRelation'.format(base_url, app_id)
        # 更新实体简介
        self.update_entity_intro_url = '{}/xeonKgDal/entityProfile/{}/addEntityProfile/{}'
        # 更新属性简介
        self.update_property_intro_url = '{}/xeonKgDal/{}/property/desc'.format(base_url, app_id)
        # 添加属性语料
        self.add_property_corpus_url = '{}/xeonKgDal/{}/property/{}'
        # 查询特殊属性
        self.load_extra_property_url = '{}/xeonKgDal/extraPropertyManager/{}/getAll?page=1&keyword={}&limit=10'
        # 更新特殊属性
        self.update_extra_property_url = '{}/xeonKgDal/extraPropertyManager/{}/update'.format(base_url, app_id)

        # 词库 新增同义词 实体/属性
        self.wordbank_add_url = '{}/xeonKgDal/synonym/{}/addSynonym'.format(base_url, app_id) + "/{}?type={}"
        # 词库 更新同义词 实体/属性
        self.wordbank_update_url = '{}/xeonKgDal/synonym/{}/updateSynonym'.format(base_url,
                                                                                  app_id) + "/{}?type={}"
        # 词库 删除同义词 实体/属性
        self.wordbank_delete_url = '{}/xeonKgDal/synonym/{}/deleteSynonym'.format(base_url,
                                                                                  app_id) + "/{}?type={}"
        # 词库 查询同义词 实体/属性
        self.wordbank_get_url = '{}/xeonKgDal/synonym/{}/showSynonym'.format(base_url,
                                                                             app_id) + "?show=matchId&type={}&object_id={}"

        # 词库 新增值同义词
        self.wordbank_value_add_url = '{}/xeonKgDal/propertyKeyword/{}/addKeywordSynonym'.format(base_url, app_id)
        # 词库 更新值同义词
        self.wordbank_value_update_url = '{}/xeonKgDal/propertyKeyword/{}/updateKeywordSynonym'.format(base_url,
                                                                                                       app_id)
        # 词库 删除同义词
        self.wordbank_value_delete_url = '{}/xeonKgDal/propertyKeyword/{}/deleteKeywordSynonym'.format(base_url,
                                                                                                       app_id)
        # 词库 查询值同义词
        self.wordbank_value_get_url = '{}/xeonKgDal/propertyKeyword/{}?target=getAll'.format(base_url, app_id)

        # 初始知识库
        self.init_url = '{}/xeonKgDal/property/{}/initialProperty'.format(base_url, app_id)

    def init_data(self):
        """
        出话kg知识库
        :param data: kg数据
        """
        try:
            response = requests.get(self.init_url)
            if self._response_failed(response):
                msg = "kg: 初始化数据失败：原因: 接口异常。" + str(response.json()) if response is not None else ''
                raise BfEngineException(ErrorCode.kg_init_error, msg)
        except BfEngineException as bfe:
            raise bfe
        except Exception:
            msg = "kg: 初始化数据失败， 原因: Exception When POST '{}' : {}".format(self.init_url, traceback.format_exc())
            raise BfEngineException(ErrorCode.kg_init_error, msg)

    def load_data(self, data: dict):
        """
        加载json格式训练数据
        :param data: kg数据
        """
        try:
            response = requests.post(self.data_load_url, json=data)
            if self._response_failed(response):
                msg = "kg: 上传训练数据失败：原因: 上传数据接口异常。" + str(response.json()) if response is not None else ''
                raise BfEngineException(ErrorCode.kg_import_error, msg)
        except BfEngineException as bfe:
            raise bfe
        except Exception:
            msg = "kg: 上传数据失败， 原因: Exception When POST '{}' : {}".format(self.data_load_url, traceback.format_exc())
            raise BfEngineException(ErrorCode.kg_import_error, msg)

    def load_file(self, file_path: str, use_old):
        """
        加载文件格式训练数据
        :param file_path: kg数据文件路径
        """
        try:
            files = {"file": (open(file_path, 'rb'))}
            file_load_url = self.new_file_load_url
            if use_old is not None and use_old:
                file_load_url = self.file_load_url
            response = requests.post(file_load_url, files=files)
            if self._response_failed(response):
                msg = "kg: 上传训练数据失败：原因: 上传数据接口异常。" + str(response.json()) if response is not None else ''
                raise BfEngineException(ErrorCode.kg_import_error, msg)
        except BfEngineException as bfe:
            raise bfe
        except Exception:
            msg = "kg: 上传数据失败， 原因: Exception When POST '{}' : {}".format(self.data_load_url, traceback.format_exc())
            raise BfEngineException(ErrorCode.kg_import_error, msg)

    def trigger_train(self):
        """
        触发训练 异步
        """
        try:
            response = requests.get(url=self.trigger_train_url)
            if self._response_failed(response):
                msg = "kg: 触发训练失败：原因: 接口异常。" + str(response.json()) if response is not None else ''
                raise BfEngineException(ErrorCode.kg_train_error, msg)
        except BfEngineException as bfe:
            raise bfe
        except Exception:
            msg = "kg: 触发训练失败， 原因: Exception When POST '{}' : {}".format(self.trigger_train_url, traceback.format_exc())
            raise BfEngineException(ErrorCode.kg_import_error, msg)

    def sync_data(self):
        """
        同步出话层数据
        """
        try:
            response = requests.get(url=self.sync_data_url)
            if self._response_failed(response):
                msg = "kg: 同步出话层数据：原因: 接口异常。" + str(response.json()) if response is not None else ''
                raise BfEngineException(ErrorCode.kg_train_error, msg)
        except BfEngineException as bfe:
            raise bfe
        except Exception:
            msg = "kg: 同步出话层数据， 原因: Exception When POST '{}' : {}".format(self.sync_data_url, traceback.format_exc())
            raise BfEngineException(ErrorCode.kg_train_error, msg)

    def sync_sandbox(self):
        """
        同步数据至线上
        """
        try:
            response = requests.post(url=self.sync_sandbox_url)
            if self._response_failed(response):
                msg = "kg: 同步数据失败：原因: 接口异常。" + str(response.json()) if response is not None else ''
                raise BfEngineException(ErrorCode.kg_train_error, msg)
        except BfEngineException as bfe:
            raise bfe
        except Exception:
            msg = "kg: 同步数据失败， 原因: Exception When POST '{}' : {}".format(self.trigger_train_url, traceback.format_exc())
            raise BfEngineException(ErrorCode.kg_train_error, msg)

    def status(self, wait_type):
        """
        等待中 并记录日志
        :param wait_type: train | sync
        :return: 是否成功
        """
        try:
            response = requests.get(url=self.train_status_url)
            if self._response_failed(response):
                msg = "kg: 获取状态失败：原因: 接口异常。" + str(response.json()) if response is not None else ''
                raise BfEngineException(ErrorCode.kg_train_error, msg)

            data = response.json().get('data')
            if data:
                train_status = data.get('trainingStatus')
                sync_status = data.get('syncStatus')
                if wait_type == 'train' and train_status == 2:
                    return True
                if wait_type == 'sync' and sync_status == 2:
                    return True
            return False
        except BfEngineException as bfe:
            raise bfe
        except Exception:
            msg = "kg: 获取状态失败， 原因: Exception When POST '{}' : {}".format(self.train_status_url, traceback.format_exc())
            raise BfEngineException(ErrorCode.kg_train_error, msg)

    def call(self, msg, online):
        """
        预测出话
        :param msg: 用户问
        :param online：线上 True 线下 False
        :return: 机器人回答
        """
        answer = Answer(text='', score=0)
        if online is None or online:
            env = "production"
        else:
            env = "sandbox"
        param = {
            "userid": "test",
            "sessionid": "test",
            "uniqueid": "test",
            "appid": self.app_id,
            "env": env,
            "modeltype": "ft",
            "text": msg,
            "skg": "on"
        }
        try:
            response = requests.post(url=self.predict_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 预测失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('response'):
                    answer = Answer(text=response_json.get('response'), score=response_json.get('score'))
                else:
                    pass
                    # logging.warning('kg: 预测为空 [{}]'.format(msg))
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.predict_url, traceback.format_exc())
            logging.error('kg: 对话失败， 原因: ' + msg)
        finally:
            if not answer.text:
                answer.value = Config.backfill_speech
                answer.score = 60
            return answer

    def check_entity_repeat(self, name):
        """
        检查实体重复
        :param name: 实体名称
        :return: 结果
        """
        try:
            response = requests.get(self.check_entity_repeat_url.format(Config.base_url, self.app_id, name))
            if self._response_failed(response):
                logging.error("kg: 检查实体重复失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data'):
                    data = response.json().get('data')
                    if data:
                        return data.get('found')
                    else:
                        return False
                else:
                    return False
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.check_entity_repeat_url, traceback.format_exc())
            logging.error('kg: 检查实体重复失败， 原因：' + msg)
            return -1

    def load_entity(self, parent_entity_id, name):
        """
        获取实体
        :param parent_entity_id: 父实体id
        :param name: 实体名称
        :return: 结果
        """
        try:
            response = requests.get(self.load_entity_url.format(Config.base_url, self.app_id, parent_entity_id, name))
            if self._response_failed(response):
                logging.error("kg: 获取实体失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('dataDetail'):
                    data = response.json().get('dataDetail')
                    if data:
                        return data[0]
                    else:
                        return None
                else:
                    return None
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.load_entity_url, traceback.format_exc())
            logging.error('kg: 获取实体失败， 原因：' + msg)
            return -1

    def add_entity(self, name):
        """
        新增一级实体
        :param name: 实体名称
        :return: 结果
        """
        param = {
            "name": name
        }
        try:
            response = requests.post(url=self.add_entity_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 新增一级实体失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data'):
                    data = response.json().get('data')
                    if data:
                        return data.get('freshEntityId')
                    else:
                        return -1
                else:
                    return -1
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.add_entity_url, traceback.format_exc())
            logging.error('kg: 新增一级实体失败， 原因：' + msg)
            return -1

    def load_property(self, name):
        """
        获取实体
        :param name: 实体名称
        :return: 结果
        """
        try:
            response = requests.get(self.load_property_url.format(Config.base_url, self.app_id, name))
            if self._response_failed(response):
                logging.error("kg: 获取属性失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data'):
                    data = response.json().get('data')
                    if data:
                        return data[0]
                    else:
                        return None
                else:
                    return None
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.load_property_url, traceback.format_exc())
            logging.error('kg: 获取属性失败， 原因：' + msg)
            return -1

    def add_property(self, name, category, unit):
        """
        新增属性
        :param name: 实体名称
        :param category: 1 entity 2 stringValue 3 integerValue(deprecated) 4 匿名实体 6 空实体 7 日期 8 浮点数
        :param unit: 单位
        :return: 结果
        """
        param = {
            "name": name,
            "corpus": [],
            "category": category,
            "unit": unit,
            "speech": "{\"answer\":\"\",\"property\":{\"dimension_id_list\":[]}}",
            "slotInfo": {
                "url": "",
                "slotList": []
            },
            "parentId": -1,
            "speechType": "201"
        }
        try:
            response = requests.post(url=self.add_property_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 新增屬性失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data'):
                    data = response.json().get('data')
                    if data:
                        return data
                    else:
                        return -1
                else:
                    return -1
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.add_property_url, traceback.format_exc())
            logging.error('kg: 新增屬性失败， 原因：' + msg)
            return -1

    def load_value(self, entity_id, property_id):
        """
        获取实体
        :entity_id: 实体id
        :property_id: 属性id
        :return: 结果
        """
        try:
            response = requests.get(self.load_relation_url.format(Config.base_url, self.app_id, entity_id))
            if self._response_failed(response):
                logging.error("kg: 获取属性失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data').get('data'):
                    data = response.json().get('data').get('data')
                    if data:
                        return data[0].get('propertyValue').get('entityPropertyValueTable')[str(entity_id)][
                            str(property_id)]
                    else:
                        return None
                else:
                    return None
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.load_property_url, traceback.format_exc())
            logging.error('kg: 获取属性失败， 原因：' + msg)
            return -1

    def add_value(self, from_entity, property, to_entity_id, to_entity_name):
        """
        新增属性
        :param name: 实体名称
        :param category: 1 entity 2 stringValue 3 integerValue(deprecated) 4 匿名实体 6 空实体 7 日期 8 浮点数
        :return: 结果
        """
        category = property.get('category')
        # if category != 1:
        #     value = "{\\\"answer\\\":\\\"" + to_entity_name + "\\\",\\\"property\\\":{\\\"dimension_id_list\\\":[]}}"
        # else:
        #     value = to_entity_name
        param = {
            "fromEntityId": from_entity.get('id'),
            "fromEntityName": from_entity.get('name'),
            "propertyId": property.get('id'),
            "category": category,
            "toEntityId": [
                to_entity_id
            ],
            "toEntityName": [
                to_entity_name
            ],
            "isInheritParent": False,
            "inheritedParentEntityId": "-1",
            "subCategory": "201"
        }

        try:
            response = requests.post(url=self.add_relation_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 新增值失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data'):
                    data = response.json().get('data')
                    if data:
                        return data.get('id')
                    else:
                        return -1
                else:
                    return -1
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.add_relation_url, traceback.format_exc())
            logging.error('kg: 新增值失败， 原因：' + msg)
            return -1

    def update_value(self, relation_id, from_entity, property, to_entity_id, to_entity_name):
        """
        新增属性
        :param name: 实体名称
        :param category: 1 entity 2 stringValue 3 integerValue(deprecated) 4 匿名实体 6 空实体 7 日期 8 浮点数
        :return: 结果
        """
        category = property['category']
        if category != 1:
            value = "{\"answer\":\"" + to_entity_name + "\",\"property\":{\"dimension_id_list\":[]}}"
        else:
            value = to_entity_name
        param = {
            "id": relation_id,
            "fromEntityId": from_entity['id'],
            "fromEntityName": from_entity['name'],
            "propertyId": property['id'],
            "propertyName": property['name'],
            "category": category,
            "unit": property['unit'],
            "toEntityId": [
                to_entity_id
            ],
            "toEntityName": [
                value
            ],
            "isInheritParent": False,
            "inheritedParentEntityId": "-1",
            "subCategory": "201"
        }

        try:
            response = requests.post(url=self.update_relation_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 更新值失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data'):
                    data = response.json().get('data')
                    if data:
                        return data.get('id')
                    else:
                        return -1
                else:
                    return -1
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.update_relation_url, traceback.format_exc())
            logging.error('kg: 更新值失败， 原因：' + msg)

    def update_entity_intro(self, entity_id, intro):
        """
        新增实体简介
        :param intro: 实体简介
        :return: 结果
        """
        param = {
            "introduction": "[{\"type\":\"INTRODUCTION\",\"entity\":[],\"property\":[],\"value\":\"{\\\"answer\\\":\\\"" + intro + "\\\",\\\"property\\\":{\\\"dimension_id_list\\\":[]}}\",\"subCategory\":\"201\"}]"
        }

        try:
            response = requests.post(url=self.update_entity_intro_url.format(Config.base_url, self.app_id, entity_id),
                                     json=param)
            if self._response_failed(response):
                logging.error("kg: 更新实体简介失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.update_entity_intro_url, traceback.format_exc())
            logging.error('kg: 更新实体简介失败， 原因：' + msg)
            return False

    def update_property_intro(self, property_id, intro):
        """
        新增属性简介
        :property_id: 属性id
        :param intro: 属性简介
        :return: 结果
        """
        param = {
            "propertyId": property_id,
            "introduction": "[{\"type\":\"INTRODUCTION\",\"entity\":[],\"property\":[],\"value\":\"{\\\"answer\\\":\\\"" + intro + "\\\",\\\"property\\\":{\\\"dimension_id_list\\\":[]}}\",\"subCategory\":\"201\"}]"
        }
        try:
            response = requests.post(url=self.update_property_intro_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 更新属性简介失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.update_property_intro_url, traceback.format_exc())
            logging.error('kg: 更新属性简介失败， 原因：' + msg)
            return False

    def add_property_corpus(self, property, corpus):
        """
        新增属性简介
        :property: 属性
        :corpus: 新增的语料
        :return: 结果
        """
        if corpus not in property.get('corpus'):
            property.get('corpus').append(corpus)
        param = {
            "name": property.get('name'),
            "corpus": property.get('corpus'),
            "category": property.get('category'),
            "unit": property.get('unit'),
            "speech": property.get('speech'),
            "parentId": -1,
            "slotInfo": {
                "url": "",
                "slotList": []
            },
            "speechType": str(property.get('speechType'))
        }
        try:
            response = requests.put(
                url=self.add_property_corpus_url.format(Config.base_url, self.app_id, property.get('id')), json=param)
            if self._response_failed(response):
                logging.error("kg: 新增属性语料失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.add_property_corpus_url, traceback.format_exc())
            logging.error('kg: 新增属性语料失败， 原因：' + msg)
            return False

    def update_property_speech(self, property, speech):
        """
        新增属性简介
        :property: 属性
        :corpus: 新增的语料
        :return: 结果
        """
        if not speech:
            return True
        param = {
            "name": property.get('name'),
            "corpus": property.get('corpus'),
            "category": property.get('category'),
            "unit": property.get('unit'),
            "speech": "{\"answer\":\"" + speech + "\",\"property\":{\"dimension_id_list\":[]}}",
            "parentId": -1,
            "slotInfo": {
                "url": "",
                "slotList": []
            },
            "speechType": str(property.get('speechType'))
        }
        try:
            response = requests.put(
                url=self.add_property_corpus_url.format(Config.base_url, self.app_id, property.get('id')), json=param)
            if self._response_failed(response):
                logging.error("kg: 更新属性话术失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.add_property_corpus_url, traceback.format_exc())
            logging.error('kg: 更新属性话术失败，原因：' + msg)
            return False

    def _get_wordbank_synonym(self, type, object_id):
        try:
            url = self.wordbank_get_url.format(type, object_id)
            response = requests.get(url=url)
            if self._response_failed(response):
                logging.error("kg: 查询同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data'):
                    return response_json.get('data')
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_get_url, traceback.format_exc())
            logging.error('kg: 查询同义词失败， 原因：' + msg)

    def add_word_synonym(self, word_name: str, synonym: list):
        """
        增加词典同义词

        :param word_name: 要添加的词条
        :param synonym: 同义词列表
        :return:
        """
        return self._add_entity_synonym(word_name, synonym) or self._add_property_synonym(word_name, synonym) or \
               self._add_value_synonym(word_name, synonym)

    def update_word_synonym(self, word_name: str, synonym: list):
        """
        更新词典同义词
        :param word_name: 要更新的词条
        :param synonym: 同义词列表
        :return:
        """
        return self._update_entity_synonym(word_name, synonym) or self._update_property_synonym(word_name, synonym) or \
               self._update_value_synonym(word_name, synonym)

    def delete_word_synonym(self, word_name: str):
        """
        删除词典词条及同义词
        :param word_name: 要删除的词条
        :return:
        """
        return self._delete_entity_synonym(word_name) or self._delete_property_synonym(word_name) or \
               self._delete_value_synonym(word_name)

    def get_word_synonym(self, word_name: str):
        """
        获取词典的同义词
        :param word_name: 要查询的词条
        :return:
        """
        return self._get_entity_synonym(word_name) or self._get_property_synonym(word_name) or \
               self._get_value_synonym(word_name)

    def _add_entity_synonym(self, entity_name: str, synonym: list):
        """
        新增实体同义词
        :param entity_name: 实体名称
        :param synonym: 实体的同义词列表
        :return:
        """
        try:
            entity_object = self.load_entity(0, entity_name)
            if entity_object is not None:
                entity_id = entity_object['id']
                entity_exist = self._get_wordbank_synonym('entity', entity_id) is not None
                if entity_exist:
                    return self._update_entity_synonym(entity_name, synonym)
                else:
                    url = self.wordbank_add_url.format(entity_id, 'entity')
                    param = {
                        "id": entity_id,
                        "synonym": synonym
                    }
                    response = requests.post(url=url, json=param)
                    if self._response_failed(response):
                        logging.error("kg: 保存实体同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
                    else:
                        return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_add_url, traceback.format_exc())
            logging.error('kg: 保存实体同义词失败， 原因：' + msg)

    def _update_entity_synonym(self, entity_name: str, synonym: list):
        """
        更新实体的同义词
        :param entity_name: 实体名称
        :param synonym: 实体同义词 列表
        :return:
        """
        try:
            entity_object = self.load_entity(0, entity_name)
            if entity_object is not None:
                entity_id = entity_object['id']
                url = self.wordbank_update_url + "/{}?type=entity".format(entity_id)
                param = {
                    "id": entity_id,
                    "synonym": synonym
                }
                response = requests.put(url=url, json=param)
                if self._response_failed(response):
                    logging.error("kg: 保存实体同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
                else:
                    return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 保存实体同义词失败， 原因：' + msg)

    def _delete_entity_synonym(self, entity_name):
        """
        删除实体的同义词
        :param entity_name: 实体名称
        :return:
        """
        try:
            entity_object = self.load_entity(0, entity_name)
            if entity_object is not None:
                entity_id = entity_object['id']
                url = self.wordbank_update_url + "/{}?type=entity".format(entity_id)
                response = requests.delete(url=url)
                if self._response_failed(response):
                    logging.error("kg: 删除实体同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
                else:
                    return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 删除实体同义词失败， 原因：' + msg)

    def _get_entity_synonym(self, entity_name):
        """
        获取实体的同义词
        :param entity_name: 实体名称
        :return:
        """
        try:
            entity_object = self.load_entity(0, entity_name)
            if entity_object is not None:
                entity_id = entity_object['id']
                return self._get_wordbank_synonym('entity', entity_id)
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 获取实体同义词失败， 原因：' + msg)

    def _add_property_synonym(self, property_name: str, synonym: list):
        """
        新增实体同义词
        :param property_name: 属性名称
        :param synonym: 属性的同义词列表
        :return:
        """
        try:
            property_object = self.load_property(property_name)
            if property_object is not None:
                property_id = property_object['id']
                property_exist = self._get_wordbank_synonym('property', property_id) is not None
                if property_exist:
                    return self._update_property_synonym(property_name, synonym)
                else:
                    url = self.wordbank_add_url.format(property_id, 'property')
                    param = {
                        "id": property_id,
                        "synonym": synonym
                    }
                    response = requests.post(url=url, json=param)
                    if self._response_failed(response):
                        logging.error("kg: 保存属性同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
                    else:
                        return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_add_url, traceback.format_exc())
            logging.error('kg: 保存属性同义词失败， 原因：' + msg)

    def _update_property_synonym(self, property_name: str, synonym: list):
        """
        更新属性的同义词
        :param property_name: 属性名称
        :param synonym: 属性同义词 列表
        :return:
        """
        try:
            property_object = self.load_property(property_name)
            if property_object is not None:
                property_id = property_object['id']
                url = self.wordbank_update_url.format(property_id, "property")
                param = {
                    "id": property_id,
                    "synonym": synonym
                }
                response = requests.put(url=url, json=param)
                if self._response_failed(response):
                    logging.error("kg: 保存属性同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
                else:
                    return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 保存属性同义词失败， 原因：' + msg)

    def _delete_property_synonym(self, property_name: str):
        """
        删除属性的同义词
        :param property_name: 属性名称
        :return:
        """
        try:
            property_object = self.load_property(property_name)
            if property_object is not None:
                property_id = property_object['id']
                url = self.wordbank_update_url.format(property_id, "property")
                response = requests.delete(url=url)
                if self._response_failed(response):
                    logging.error("kg: 删除属性同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
                else:
                    return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 删除属性同义词失败， 原因：' + msg)

    def _get_property_synonym(self, property_name: str):
        """
        获取属性的同义词
        :param property_name: 属性名称
        :return:
        """
        try:
            property_object = self.load_property(property_name)
            if property_object is not None:
                property_id = property_object['id']
                return self._get_wordbank_synonym("property", property_id)
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 获取属性同义词失败， 原因：' + msg)

    def _add_value_synonym(self, value_name: str, synonym: list):
        """
        增加值的同义词
        :param value_name: 值名称
        :param synonym: 值的同义词
        :return:
        """
        try:
            value_object = self._get_value_synonym(value_name)
            if value_object is not None:
                self._update_value_synonym(value_name, synonym)
            else:
                param = {
                    'name': value_name,
                    'synonym': synonym
                }

                response = requests.post(url=self.wordbank_value_add_url, json=param)
                if self._response_failed(response):
                    logging.error("kg: 新增值同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
                else:
                    return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 删除属性同义词失败， 原因：' + msg)

    def _get_value_synonym(self, value_name: str):
        """
        获取值的同义词
        :param value_name: 值名称
        :return:
        """
        try:
            response = requests.get(url=self.wordbank_value_get_url)
            if self._response_failed(response):
                logging.error("kg: 查询值同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
            else:
                data = response.json().get('data')
                if data and type(data) == 'list':
                    filter_objects = list(filter(lambda x: x['name'] == value_name, data))
                    if len(filter_objects) > 0:
                        return filter_objects[0]
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 查询值同义词失败， 原因：' + msg)

    def _update_value_synonym(self, value_name: str, synonym: list):
        """
        更新值的同义词
        :param value_name: 值名称
        :param synonym: 值的同义词 列表
        :return:
        """
        try:
            param = {
                'name': value_name,
                'synonym': synonym
            }
            response = requests.post(url=self.wordbank_value_update_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 更新值同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
            else:
                return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 更新值同义词失败， 原因：' + msg)

    def _delete_value_synonym(self, value_name: str):
        """
        删除值的同义词
        :param value_name:  值名称
        :return:
        """
        try:
            param = {
                'name': value_name,
            }
            response = requests.delete(url=self.wordbank_value_update_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 删除值同义词失败：原因: 接口异常。" + str(response.text) if response is not None else '')
            else:
                return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.wordbank_update_url, traceback.format_exc())
            logging.error('kg: 删除值同义词失败， 原因：' + msg)

    def load_extra_property(self, property_name, entity_name):
        """
        获取实体
        :property_name: 属性名称
        :entity_name: 实体名称
        :return: 结果
        """
        try:
            response = requests.get(self.load_extra_property_url.format(Config.base_url, self.app_id, property_name))
            if self._response_failed(response):
                logging.error("kg: 获取特殊属性失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data').get('result'):
                    results = response.json().get('data').get('result')
                    if results:
                        for result in results:
                            if result.get('entityName') == entity_name:
                                return result
                        return None
                    else:
                        return None
                else:
                    return None
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.load_property_url, traceback.format_exc())
            logging.error('kg: 获取特殊属性失败， 原因：' + msg)
            return None

    def add_extra_property(self, property_id, entity_id, category, unit):
        """
        获取实体
        :property_id: 属性id
        :entity_id: 实体id
        :category: 属性类型
        :unit: 单位
        :return: 结果
        """
        param = {
            "propertyId": property_id,
            "extraProperties": [{
                "entityId": entity_id,
                "category": category,
                "unit": unit,
                "speech": "{\"answer\":\"\",\"property\":{\"dimension_id_list\":[]}}",
                "speechType": "201",
                "extra_content": {}
            }]
        }
        try:
            response = requests.post(url=self.update_extra_property_url, json=param)
            if self._response_failed(response):
                logging.error("kg: 更新特殊属性失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.update_extra_property_url, traceback.format_exc())
            logging.error('kg: 更新特殊属性失败， 原因：' + msg)
            return False

    def update_extra_property(self, property_name, entity_name, category, unit):
        """
        获取实体
        :property_name: 属性名称
        :entity_name: 实体名称
        :return: 结果
        """
        param = {
            "propertyId": 0,
            "extraProperties": [{
                "entityId": 0,
                "category": category,
                "unit": unit,
                "speech": "{\"answer\":\"\",\"property\":{\"dimension_id_list\":[]}}",
                "speechType": "201",
                "extra_content": {}
            }]
        }
        try:
            response = requests.get(self.load_extra_property_url.format(Config.base_url, self.app_id, property_name))
            if self._response_failed(response):
                logging.error("kg: 获取特殊属性失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data').get('result'):
                    results = response.json().get('data').get('result')
                    if results:
                        for result in results:
                            if result.get('entityName') == entity_name:
                                result['category'] = category
                                result['unit'] = unit
                                param['propertyId'] = result.get('propertyId')
                                result['extra_content'] = result['extraContent']
                                param['extraProperties'] = result
                                response = requests.post(url=self.update_extra_property_url, json=param)
                                if self._response_failed(response):
                                    logging.error(
                                        "kg: 更新特殊属性失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
                                    return False
                return True
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.load_property_url, traceback.format_exc())
            logging.error('kg: 获取特殊属性失败， 原因：' + msg)
            return False



import time
import traceback
import os
import xlsxwriter
import xlrd

from tqdm import tqdm

from .module import Module
from ..caller.qa_caller import QACaller
from ..caller.intent_caller import  IntentCaller
from ..entity.answer import Answer
from ..entity.enums import QuestionType, QuestionMode,QuestionField
from ..entity.exception import BfEngineException, ErrorCode
from ..logger import log


class IntentAnswering(Module):
    """
    问答
    """
    tmp_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
    faq_sq_empty_path = tmp_dir + "/data/问答上传模板.xlsx"
    faq_lq_empty_path = tmp_dir + "/data/语料上传模板.xlsx"

    def __init__(self, app_id, set, init):
        super().__init__(app_id, 'qa', set)
        self.intentCaller = IntentCaller(app_id)
        if init:
            self.init()

    def train(self, data: str = None,intent_path: str = None,append: bool = False):
        """
        :param data:
        :param append: 是否追加，True
        :return: 训练是否成功
        """
        try:
            log.info('intent: prepare train')
            if data:
                group_ids = []
                for item in data["data"]:
                    try:
                        group_id = self._add_group(item["domain"])
                        group_ids.append(group_id)
                        self._add_intents(group_id=group_id, intents=item["intents"])
                    except BfEngineException as bfe:
                        log.error(bfe)
                for group_id in group_ids:
                    self._train(group_id=group_id)
            elif intent_path:
                return
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error('unknown exception')
            log.error(e)
            return False
        return True
    def open_inner_intents(self):
        self.intentCaller.open_default_intent()

    def close_inner_intents(self):
        self.intentCaller.close_default_intent()

    def close_intent_group(self,group_id:str=None):
        self.intentCaller.close_intent_group(group_id=group_id)

    def open_intent_group(self, group_id: str = None):
        self.intentCaller.open_intent_group(group_id=group_id)
    def predict(self, text: str,online: bool=False) -> Answer:
        data = self.intentCaller.predict(text,online)
        if not data:
            return None
        if len(data)==0:
            return None
        if float(data[0]["score"])<float(self.set.intent_threshold):
            return None
        return Answer(text=data[0]["label"],score=data[0]["score"])

    def intents_list(self):
        return self._intents_list()

    def _add_group(self,domain:str = None):
        return self.intentCaller.add_group(domain=domain)
    def _add_intents(self,group_id:str=None,intents:str=None):
        for intent in intents:
            self._add_intent(group_id=group_id,intent=intent)
    def _add_intent(self,group_id:str=None,intent:str=None):
        return self.intentCaller.add_intent(group_id=group_id,intent=intent)
    def _train(self,group_id:str=None,):
        log.info('intent: start training')
        train_id = self.intentCaller.train(group_id=group_id)
        progress = 0
        with tqdm(total=100) as bar:
            bar.set_description("intent training...")
            if train_id == 0:
                bar.update(100)
                bar.set_description("intent train finished")
                return
            while progress < 100:
                old_progress = progress
                progress = self.intentCaller.train_status(group_id=group_id)
                bar.update(progress - old_progress)
                time.sleep(10)
            bar.set_description("intent train finished")
        time.sleep(10)
    def _intents_list(self):
        return self.intentCaller.intent_group_list()
    def _export(self,sq_path: str = None, lq_path: str = None):
        """
        导出问答和语料文件
        :param sq_path 问答文件路径
        :param lq_path 语料文件路径
        """
        try:
            if sq_path:
                data_path = self.qaCaller.download_launch(QuestionType.SQ_ANS)
                self.qaCaller.download(sq_path,data_path)
            if lq_path:
                data_path = self.qaCaller.download_launch(QuestionType.LQ)
                self.qaCaller.download(lq_path, data_path)
        except BfEngineException as bfe:
            log.error(bfe)
        except Exception as e:
            log.error(e)
    def _read_intent_path(self,intent_path:str=None)->list:
        workbook = xlrd.open_workbook(intent_path)
        if workbook.sheets()==0:
            return []
        sheet = workbook.sheet_by_index(0)
        intent = {}
        for row in range(1, sheet.nrows):  # 行
            name = sheet.cell_value(row, 0)  # 语料
            corpus = sheet.cell_value(row, 1) # 意图名称
            if name in intent.keys():
                intent[name].append(corpus)
            else:
                intent[name]=[corpus]
        return intent


import requests

from ..logger import log
from .base import CallerBase
from ..config import Config
from ..entity.enums import QuestionType, QuestionMode, QuestionField
from ..entity.exception import BfEngineException
from .set_config import SetConfig
from .langconv import Converter
import random

class QACaller(CallerBase):
    """
    QA api调用
    """

    def __init__(self, app_id):
        super().__init__(app_id)
        self.app_id = app_id
        self.module = 'qa'
        self.header = {
            "X-locale": "zh-cn",
            "app_id": app_id,
            "user_id": "bf-engine-sdk",
            "Authorization": "Bearer EMOTIBOTDEBUGGER",
            "Accept": "application/json,text/plain, */*"
        }
        self.upload_faq_upload_url = Config.base_url + "/ssm/dac/upload"
        self.upload_faq_upload_status_url = Config.base_url + "/ssm/dac/upload/"
        self.upload_faq_train_url = Config.base_url + "/ssm/dac/train"
        self.upload_faq_train_status_url = Config.base_url + "/ssm/dac/trainhistory"
        self.upload_faq_json_sq_url = Config.base_url + "/ssm/dac/sq"
        self.upload_faq_json_lq_url = Config.base_url + "/ssm/dac/lq"
        self.upload_faq_json_answer_url = Config.base_url + "/ssm/dac/answer"
        self.upload_faq_json_publish_url = Config.base_url + "/ssm/dac/release/online"
        self.upload_faq_json_publish_status_url = Config.base_url + "/ssm/dac/release/progress"
        self.upload_faq_json_publish_status_url = Config.base_url + "/ssm/dac/release/progress"
        self.upload_faq_json_tag_url = Config.base_url + "/ssm/dac/tag"
        self.download_faq_download_url = Config.base_url + "/ssm/dac/download"
        self.download_faq_download_export_url = Config.base_url + "/ssm/dac/common/minio/excelfile"
        self.search_faq_json_question_list_url = Config.base_url + "/ssm/dac/info"
        self.search_faq_json_corpus_list_url = Config.base_url + "/ssm/dac/lq"
        self.search_faq_json_related_sq_list_url = Config.base_url + "/ssm/dac/sq/related_candidate"
        self.operate_faq_category_url  = Config.base_url + "/ssm/dac/category"
        self.get_faq_sub_category_url  = Config.base_url + "/ssm/dac/category/sub"
        self.get_predict_qa_list_url = Config.base_url + "/qa/predict"

    def upload(self, data_type: QuestionType, data_model: QuestionMode, file_path: str) -> str:
        """
        上传标注问/语料
        :param data_type: 上传类型 (SQ_ANS|LQ|TQ)
        :param data_model: 上传模型 (全量|增量)
        :param file_path: 上传文件路径
        :return 上传id
        """
        # 上传问题$答案
        files = {"file": (self.app_id + ".xlsx", open(file_path, 'rb'))}
        data = {"type": data_type, "mode": data_model, "comment": "BFEngine 导入 qa"}
        resp = requests.request("POST", self.upload_faq_upload_url, headers=self.header, data=data, files=files).json()

        # 问题$答案上传进度
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return str(resp["data"])

    def upload_status(self, data_id: str, is_log: bool = True) -> int:
        """
        查询上传状态
        :param is_log: 是否打日志
        :param data_id: 上传id
        :return 上传进度
        """
        resp = requests.get(self.upload_faq_upload_status_url + "/" + data_id, headers=self.header, json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        progress = resp["data"]["progress"]
        # if is_log:
        #     log.info(self.module + ": 上传进度==" + str(progress) + "%")
        return progress

    def download_launch(self, data_type: QuestionType) -> str:
        """
        发起导出标注问/语料
        :param data_type: 下载类型 (SQ_ANS|LQ|TQ)
        :param file_path: 下载文件路径
        :return 服务器文件下载路径
        """
        resp = requests.get(self.download_faq_download_url + "?type=" + data_type, headers=self.header, json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["data"]

    def download(self, file_path: str, service_file_path: str, is_log: bool = True):
        """
        从服务获取文件
        :param is_log: 是否打日志
        :param file_path: 本地文件保存路径
        :param data_path: 服务文件路径
        :return 上传进度
        """
        resp = requests.get(self.download_faq_download_export_url + "?filename=" + service_file_path, stream=True)
        with open(file_path, "wb") as f:
            for chunk in resp.iter_content(chunk_size=512):
                f.write(chunk)

    def upload_json_sq(self, tag: SetConfig = SetConfig(), data: dict = {}) -> str:
        """
        上传标准问（新增标准问）
        :param tag:  标签
        :param data: 上传数据
        :return id
        """
        tag_id_list = []
        if "tags" in data.keys():
            temps = self._tags_add(data["tags"])
            for temp in temps:
                tag_id_list.append(temp)
        related_sq_id_list = []
        if "related" in data.keys():
            temps = self._related_sq_search(data["related"])
            for temp in temps:
                related_sq_id_list.append(temp["id"])
        answers = []
        if "answers" in data.keys():
            for answer in data["answers"]:
                if "time_tag" in answer.keys():
                    if not tag.exits(answer["time_tag"]):
                        answer.pop("time_tag")
                        answers.append(answer)
                        continue
                    answer.update(tag.get(answer["time_tag"]))
                    answer.pop("time_tag")
                    answers.append(answer)
                else:
                    answers.append(answer)
        else:
            answers.append(
                {
                    "answer": data["answer"],
                    "property": {
                        "dimension_id_list": []
                    },
                    "start_time": "",
                    "end_time": "",
                    "period_type": 0,
                    "related_sq_id_list": related_sq_id_list
                }
            )
        data = {
            "sq": data["sq"],
            "category_id": int(data["category_id"]),
            "tag_id_list": tag_id_list,
            "answers": answers
        }
        resp = requests.post(self.upload_faq_json_sq_url, headers=self.header, json=data).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["data"]["id"]

    def upload_json_lq(self, data_id: str, data: dict) -> bool:
        """
        上传标准问语料
        :param data_id: 上传id
        :return 上传进度
        """
        data = {
            "records": [
                {
                    "sq_id": data_id,
                    "lq_list": list(map(lambda lq: {'lq': lq}, data["lq"] if "lq" in data.keys() else []))
                }
            ]
        }
        resp = requests.post(self.upload_faq_json_lq_url, headers=self.header, json=data).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return True

    def train(self):
        """
        训练标准问题
        :return 训练id
        """
        resp = requests.get(self.upload_faq_train_url, headers=self.header, json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return str(resp["data"])

    def train_status(self, train_id):
        """
        查询训练状态
        :param train_id: 训练id
        :return 训练进度
        """
        resp = requests.get(self.upload_faq_train_status_url + "/" + train_id, headers=self.header, json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        progress = resp["data"]["progress"]
        # log.info(self.module + ": training..." + str(progress) + "%")
        return progress

    def publish(self):
        """
        开始发布
        :return 发布id
        """
        resp = requests.get(self.upload_faq_json_publish_url, headers=self.header, json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return str(resp["data"])

    def publish_status(self, publish_id) -> int:
        """
        查询发布状态
        :param publish_id: 发布id
        :return 发布进度
        """
        resp = requests.get(self.upload_faq_json_publish_status_url + "/" + publish_id, headers=self.header,
                            json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        progress = int(resp["data"])
        return progress

    def qa_tag_add(self, name: str = None):
        """
        添加qa标签
        :param name 标签名称
        :return 标签ID
        """
        data = {"name": name}
        resp = requests.post(self.upload_faq_json_tag_url, headers=self.header,
                             json=data).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["data"]["id"]

    def qa_tag_list(self):
        """
        获取标签列表
        :return 标签列表
        """
        resp = requests.get(self.upload_faq_json_tag_url, headers=self.header,
                            json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["data"]

    def qa_list(self, field: QuestionField) -> str:
        data = self.sq_list(field)
        return data

    def sq_list(self, field: QuestionField) -> str:
        """
        标准问列表查询
        :return 标准问列表
        """
        data = {"search": [{"keyword": "", "field": str(field)}]}
        resp = requests.post(self.search_faq_json_question_list_url, headers=self.header, json=data).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        for sq_data in resp["data"]:
            sq_data['id'] = str(sq_data['id'])
            sq_data['category_id'] = str(sq_data['category_id'])
            for answer in sq_data['answers']:
                answer['id'] = str(answer['id'])
                answer['sq_id'] = str(answer['sq_id'])

        return resp["data"]

    def lq_list(self, qa_id: str):
        """
        查询标准问下语料
        :return 语料列表
        """
        resp = requests.get(self.search_faq_json_corpus_list_url + "/" + qa_id, headers=self.header,
                            json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["data"]["list"]

    def _related_sq_search(self, relates):
        """
        获取相关问列表
        :return 语料列表
        """
        related_list = []
        resp = requests.get(self.search_faq_json_related_sq_list_url, headers=self.header,
                            json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        data_list = resp["data"]
        for data in data_list:
            if data["content"] in relates:
                related_list.append(data)
        return related_list

    def _tags_add(self, tags):
        """
        添加标签集
        :return 返回添加后的标签ID集
        """
        tag_ids = []
        exist_tag_name_list = []  # 已存在标签名称
        tmp = list(set(tags))
        for tag in tmp:
            try:
                tag_ids.append(self.qa_tag_add(tag))
            except BfEngineException as bfe:
                exist_tag_name_list.append(tag)
        exist_tag_list = self._qa_tag_search(exist_tag_name_list)
        for exist_tag in exist_tag_list:
            if "id" in exist_tag.keys():
                tag_ids.append(exist_tag["id"])
        return tag_ids

    def _qa_tag_search(self, tags):
        """
        查询已经存在标签的IDS
        :param 查询tags
        :return 标签列表
        """
        tag_objs = []
        tag_obj_list = self.qa_tag_list()
        for tag_name in tags:
            for tag_obj in tag_obj_list:
                if tag_name == tag_obj["name"]:
                    tag_objs.append(tag_obj)
        return tag_objs

    def update_corpus(self, question_id, corpus_id, corpus):
        """
        更新单条语料
        :return:
        """
        params = {
            'records': [
                {
                    'sq_id': int(question_id),
                    'lq_list': [
                        {
                            'id': int(corpus_id),
                            'lq': corpus
                        }
                    ]
                }
            ]
        }
        resp = requests.put(self.upload_faq_json_lq_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0 and msg!="语料重复":
            raise BfEngineException(code=code, msg=msg)
        return True

    def add_corpus(self, question_id, corpus):
        """
        增加语料
        :return:
        """
        params = {
            'records': [
                {
                    'sq_id': int(question_id),
                    'lq_list': [
                        {
                            'lq': corpus
                        }
                    ]
                }
            ]
        }
        resp = requests.post(self.upload_faq_json_lq_url, headers=self.header, json=params, timeout=30).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        lq_id = ""
        lq_list = self.lq_list(question_id)
        for lq in lq_list:
            if lq['content'] == str(corpus):
                lq_id = lq['id']
        return lq_id

    def add_corpus_list(self, question_id, corpus_list):
        """
        增加语料
        :return:
        """
        params = {
            'records': [
                {
                    'sq_id': int(question_id),
                    'lq_list': corpus_list
                }
            ]
        }
        resp = requests.post(self.upload_faq_json_lq_url, headers=self.header, json=params, timeout=30).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return True

    def delete_corpus(self,corpus_id):
        """
        删除语料
        :return:
        """
        params = {
            'records': [
                {
                    'lq_list': [
                        {
                            'id': int(corpus_id)
                        }
                    ]
                }
            ]
        }
        resp = requests.delete(self.upload_faq_json_lq_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return True

    def delete_corpus_list(self, question_id, corpus_id_list):
        """
        删除语料
        :return:
        """
        params = {
            'records': [
                {
                    'sq_id': int(question_id),
                    'lq_list': corpus_id_list
                }
            ]
        }
        resp = requests.delete(self.upload_faq_json_lq_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return True

    def query_corpus(self, question_id, keyword=None, page=None, size=None):
        url = self.upload_faq_json_lq_url + '/' + str(question_id)
        if keyword:
            url = url + "?keyword=" + keyword.strip()
        if page is not None and size is not None:
            url = url + "&page_size=" + size + "&page_index=" + page
        resp = requests.get(url, headers=self.header, json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["data"]["list"]

    def update_question(self, data: dict) -> str:
        """
        更新问题
        :param question:
        :return:
        """
        ## 问题调更新接口，语料删除重插
        question_id = data.get('id')
        data["id"] = int(data["id"])
        if( "tag_id_list" not in data.keys()):
            data.update({"tag_id_list": []})
        if("category_id" in data.keys()):
             data["category_id"] = int(data["category_id"])
        else:
             data.update({"category_id": -1})
        requests.put(self.upload_faq_json_sq_url, headers=self.header, json=data).json()
        if("lq" in data.keys()):
            lq_request_list= data.get("lq")
            if(len(lq_request_list) > 0):
               lq_list = self.query_corpus(question_id)
               lq_id_list = []
               for lq in lq_list:
                  lq_id_list.append({"id":lq['id']})
               self.delete_corpus_list(question_id,lq_id_list)
               lq_insert_list=[]
               for lq in lq_request_list:
                   lq_insert_list.append({"lq":lq})
               self.add_corpus_list(question_id,lq_insert_list)
        if("answers" in data.keys()):
             answer_request_list = data.get("answers")
             if(len(answer_request_list) > 0):
               ans_resp = requests.get(self.upload_faq_json_answer_url + "/" + str(question_id), headers=self.header, json={}).json()
               code = int(ans_resp["code"])
               msg = str(ans_resp["message"])
               if code != 0:
                 raise BfEngineException(code=code, msg=msg)
               answer_id_list= []
               for ans in ans_resp['data']['answers']:
                   answer_id_list.append(ans['id'])
               self.delete_answer_list(answer_id_list)
               for insert_answer in answer_request_list:
                   insert_data = {
                       "answer":insert_answer,
                       "start_time":"",
                       "end_time":"",
                       "sq_id":int(question_id),
                       "property":{
                           "dimension_id_list":[

                           ]
                       },
                       "tag_id_list": [],
                       "related_sq_id_list":[]
                   }
                   self.add_answer(insert_data)
        return question_id

    def delete_question(self, question_id):
        """
        删除问题
        :param question:
        :return:
        """

        params = {
            'records': [
                int(question_id)
            ]
        }
        resp = requests.delete(self.upload_faq_json_sq_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return True

    def add_answer(self, data: dict):
        """
        新增答案
        :return:
        """
        resp = requests.post(self.upload_faq_json_answer_url, headers=self.header, json=data).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp['data']['id']

    def update_answer(self, data: dict):
        """
        更新答案
        :return:
        """
        """
        新增答案
        :return:
        """
        data["id"] = int(data["id"])
        data["sq_id"] = int(data["sq_id"])
        resp = requests.put(self.upload_faq_json_answer_url, headers=self.header, json=data).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return True

    def delete_answer(self, answer_id):
        """
        删除答案
        :return:
        """
        params = {
            'records': [answer_id]
        }
        resp = requests.delete(self.upload_faq_json_answer_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return

    def delete_answer_list(self, answer_id_list):
        """
        删除答案
        :return:
        """
        params = {
            'records': answer_id_list
        }
        resp = requests.delete(self.upload_faq_json_answer_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return

    def query_category_list(self):
        resp = requests.get(self.operate_faq_category_url, headers=self.header,
                            json={}).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["data"]

    def add_category(self,parent_id,name:str):
        params = {
            'parent_id': int(parent_id),
            'name': name
        }
        resp = requests.post(self.operate_faq_category_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp['data']['id']

    def update_category(self,id,name:str):
        params = {
            'id': int(id),
            'name': name
        }
        resp = requests.put(self.operate_faq_category_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp['data']['id']

    def delete_category(self,id):
        params = {
            'id': int(id)
        }
        resp = requests.delete(self.operate_faq_category_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return

    def get_similar_recomend_qa_list(self,user_inputs:str,type:str):
        sq_result = []
        converter = Converter('zh-hans')
        params = {"Text":converter.convert(user_inputs).lower(),'UniqueId':str(random.randint(100,10000)),"IsRelease": False,'Robot':self.app_id}
        print("params: "+str(params))
        print("predict request url: "+self.get_predict_qa_list_url)
        resp = requests.request("POST", self.get_predict_qa_list_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        print(resp['data'])
        if code != 0:
           raise BfEngineException(code=code, msg=msg)
        change_range = 0
        if(type == "similar"):
           change_range = 3
        elif(type == "recommend"):
             change_range = 25

        for match_sq in resp['data']:
            if match_sq["score"] > match_sq["threshold"]-change_range and len(sq_result)<3:
                sq_result.append(match_sq)
        return sq_result

    def get_similar_qa_list(self,user_inputs:str,max_topn,category_ids,online: bool=True):
        similar_sq_result = []
        sq_list = [];
        sq_map = {};
        answer_map = {};
        log.info(self.module + "开始")
        category_ids_list = self.get_all_sub_list(category_ids)
        log.info(self.module + "子目录查询完毕")
        self.get_all_sq_under_input_folders(category_ids_list, sq_list, sq_map, answer_map)
        log.info(self.module + "sq获取完毕")
        converter = Converter('zh-hans')
        params = {"Text":converter.convert(user_inputs).lower(),'UniqueId':str(random.randint(100,10000)),"IsRelease": online,'Robot':self.app_id}
        log.info("params: "+str(params))
        log.info("predict request url: "+self.get_predict_qa_list_url)
        resp = requests.request("POST", self.get_predict_qa_list_url, headers=self.header, json=params).json()
        code = int(resp["code"])
        msg = str(resp["message"])
        print(resp['data'])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        log.info(self.module + "预测结果获取完毕")
        for match_sq in resp['data']:
            if 'sqId' in match_sq and match_sq['sqId'] in sq_list:
               match_sq['category_id'] = str(sq_map.get(match_sq['sqId']))
               similar_sq_result.append(match_sq)
        log.info(self.module + "结果处理完毕")
        topn = int(max_topn)
        similar_sq_result = sorted(similar_sq_result, key=lambda x: x["score"], reverse=True)[:topn]
        for result_sq in similar_sq_result:
            ans_list = [];
            ans_resp_list = [];
            ans_resp_list = answer_map.get(result_sq['sqId'])
            for ans in ans_resp_list:
                ans_list.append(ans['answer'])
            result_sq['answer'] = ans_list
        log.info(self.module + "返回结果")
        return similar_sq_result

    def get_all_sq_under_input_folders(self, category_ids_list, sq_list, sq_map,answer_map):
        data = {"search": [{"keyword": "", "field": "sq"}],
                "conditions": {"category": {"id_list": category_ids_list, "logical_op": "and"}}}
        sq_resp = requests.post(self.search_faq_json_question_list_url, headers=self.header, json=data).json()
        code = int(sq_resp["code"])
        msg = str(sq_resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        for sq in sq_resp['data']:
            sq_list.append(sq['id'])
            kv = {sq['id']: sq['category_id']}
            sq_map.update(kv)
            kvanswer = {sq['id']: sq['answers']}
            answer_map.update(kvanswer)

    def get_all_sub_list(self, category_ids):
        category_ids_list = []
        for category_id in category_ids:
            header = self.header
            header.update({"folder_id": str(category_id)})
            result = requests.get(self.get_faq_sub_category_url, headers=header, json={}).json()
            category_sub_ids_list = result["data"]["idList"]
            for category_sub_id in category_sub_ids_list:
                if not category_ids_list.__contains__(category_sub_id):
                    category_ids_list.append(category_sub_id)
        return category_ids_list







class ErrorCode:
    faq_import_error = 10010
    faq_train_error = 10011
    kg_import_error = 10020
    kg_train_error = 10021
    kg_init_error = 10022
    dm_load_error = 10001
    argument_missing_error = 20001


class BfEngineException(Exception):
    def __init__(self, code, msg):
        self.code = code
        self.message = msg

    def __str__(self):
        return '{}: {}'.format(self.code, self.message)



import json

import requests

from ..entity.answer import Answer
from ..config import Config


class CallerBase:
    """
    api调用
    """

    def __init__(self, app_id):
        self.app_id = app_id
        self.base_url = Config.base_url
        self.robot_url = '{}/{}'.format(Config.base_url, Config.robot_url)

    def call_module(self, module: str, text: str, online: bool = False) -> Answer:
        """
        robot接口访问
        :param module: 模块名称
        :param text: 用户问
        :return: 机器人回答
        """
        header = {
            "Content-Type": "application/json"
        }
        data = {
            "user_id": 'bf-engine',
            "method": "inquiry",
            "app_id": self.app_id,
            "question": text,
            "included_modules": [module],
            "request_data": {
                "online": online
            }
        }
        resp = requests.post(self.robot_url, headers=header, json=data)
        params = json.loads(resp.text)
        results = params['results']
        if results:
            result = results[0]
            text = result['answer'][0]['value']
            score = result['answer_score']
            answer = Answer(text=text, score=score)
            if len(result["answer"]) == 1:
                return answer
            answer_data = result['answer'][1]["data"]
            answer.related = answer_data if answer_data else []
            return answer
        return Answer(text=Config.backfill_speech, score=60)

    @staticmethod
    def _response_failed(response):
        return response is None or response.status_code != 200 or response.json() is None



import logging
import traceback

import requests

from ..config import Config
from .base import CallerBase
from .. import util

backfill_answer_type = 3


class ConfigManagerCaller(CallerBase):

    def __init__(self, app_id):
        super().__init__(app_id)
        self.app_id = app_id
        # 获取预设话术url
        self.get_default_response_url = "{}/api/v2/robot/chats".format(Config.base_url)
        self.backfill_answer_url = "{}/api/v2/robot/chat/3/content".format(Config.base_url)
        self.create_bot_url = "{}/auth/v6/enterprise/bb3e3925f0ad11e7bd860242ac120003/app".format(Config.base_url)
        self.header = {
            "X-locale": "zh-cn",
            "app_id": self.app_id,
            "X-Appid": self.app_id,
            "user_id": "bf-engine-sdk",
            "Authorization": "Bearer EMOTIBOTDEBUGGER",
            "Accept": "application/json,text/plain, */*"
        }

    def init_bot_config(self):
        try:
            response = requests.post(url=self.create_bot_url, headers=self.header, files={
                'name': (None, util.random_uuid()),
                'description': (None, ''),
                'app_type': (None, '0'),
                'props': (None, '[{"id":1},{"id":3}]'),
                'appId': (None, self.app_id)
            })
            if response.status_code == 400:
                pass
            elif self._response_failed(response):
                logging.error(
                    "configManager: 初始化机器人失败: 接口异常。" + str(response.json()) if response is not None else '')
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.get_default_response_url, traceback.format_exc())
            logging.error('configManager: 初始化机器人失败， 原因：' + msg)

    def get_backfill(self):
        try:
            response = requests.get(url=self.get_default_response_url, headers=self.header)
            if self._response_failed(response):
                logging.error(
                    "configManager: 获取预设未知回复话术失败: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                backfill_answers = []
                response_json = response.json()
                all_default_answers = response_json['result']
                filters_backfill_answers = list(
                    filter(lambda x: x['type'] == backfill_answer_type, all_default_answers))
                if len(filters_backfill_answers) > 0:
                    backfill_answers = filters_backfill_answers[0]['contents']
                return backfill_answers
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.get_default_response_url, traceback.format_exc())
            logging.error('configManager: 获取预设未知回复话术失败， 原因：' + msg)

    def set_backfill(self, backfill_answers: list) -> bool:
        try:
            original_answers = self.get_backfill()
            for answer in original_answers:
                requests.delete(self.backfill_answer_url + "/" + str(answer['id']), headers=self.header)
            for answer in backfill_answers:
                requests.post(self.backfill_answer_url, headers=self.header, files={
                    'content': (None, answer)
                })
            return True
        except Exception:
            msg = "Exception When GET '{}' : {}".format(self.get_default_response_url, traceback.format_exc())
            logging.error('configManager: 设置未知回复话术失败， 原因：' + msg)
            return False



from .module import Module
from ..caller.parser_caller import NERParserCaller


class PredictAnswer:
    def __init__(self, slot_content, slot_name, slot_desc):
        """
        :param slot_content: 槽位预测内容
        :param slot_name: 槽位标签名称
        :param slot_desc: 槽位描述
        """
        self.slot_content = slot_content
        self.slot_name = slot_name
        self.slot_desc = slot_desc


class Parser:
    def __init__(self, name, parserId, desc, slots, userId):
        """
        解析器信息
        :param name: 解析器名称
        :param parserId: 解析器调用ID
        :param desc: 解析器用途描述
        :param slots: 解析器所能抽取出的槽位列表 list(ParserSlot)
        :param userId: 解析器所属的用户id
        """
        self.name = name
        self.parserId = parserId
        self.desc = desc
        self.userId = userId
        self.slots = slots


class ParserSlot:
    def __init__(self, name, desc):
        """
        解析器抽取的槽位信息
        :param name: 槽位名称
        :param desc: 槽位的描述
        """
        self.name = name
        self.desc = desc


class NERParser(Module):

    def __init__(self, app_id, set, init):
        super().__init__(app_id, 'ner', set)
        self.app_id = app_id
        self.caller = NERParserCaller(app_id)
        if init:
            self.init()

    def init(self):
        self.caller.register_tde_user(self.app_id)
        sys_parsers = {parser['parserId'].replace('@1', ''): parser['parserId'] for parser in
                       self.caller.get_parser_library()}
        self.sys_parser_id_map = sys_parsers

    def predict(self, sentence, parsers: list):
        """
        解析器预测
        :param sentence: 用户输入
        :param parsers: 需要使用的解析器列表
        :return:  list(PredictAnswer)
        """
        answers = []
        for parser_id in parsers or []:
            parser_id = self.sys_parser_id_map[parser_id] if parser_id in self.sys_parser_id_map else parser_id
            results = self.caller.predict(sentence, parser_id)
            if results is not None:
                results = [PredictAnswer(result['slotContent'], result['slotName'], result['slotDesc']) for result in
                           results]
                answers.extend(results)

        return answers

    def get_parsers(self):
        """
        获取机器人能调用的所有parser
        :return: list(Parser)
        """
        # todo custom parser
        user_parsers = self.caller.get_parser_library() or []
        parsers = []
        for parser in user_parsers:
            slots = []
            for slot in parser.get('slotInfo') or []:
                slots.append(ParserSlot(slot['slotName'], slot['slotDesc']))
            parsers.append(Parser(parser['name'], parser['parserId'].replace('@1', ''), parser['description'], slots,
                                  parser['userId']))

        return parsers



import logging
import traceback

import requests

from .base import CallerBase


class DialogueActCaller(CallerBase):
    """
    对话行为分类器API
    """

    def __init__(self, app_id):
        super().__init__(app_id)
        self.app_id = app_id
        self.module = 'ner'
        self.act_predict_all_url = self.base_url + "/dialogActClassify/customClassifier/detectSystemActs"
        self.act_predict_url = self.base_url + "/tde/custtagger/ud_sys_dialog_act"
        self.sys_act_get_url = self.base_url + "/tde/custtagger/sys_dialog_acts"

    def _predict_all(self, sentence: str):
        params = {
            "useId": self.app_id,
            "query": sentence
        }
        try:
            response = requests.post(url=self.act_predict_all_url, json=params)
            if self._response_failed(response):
                logging.error("act: 调用对话行为分类器失败: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                acts = response.json()['data']['sysActs']
                return [
                    {
                        'act': act['sysActLogicName'], 'score': act['confidence']}
                    for act in acts
                ]
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.act_predict_all_url, traceback.format_exc())
            logging.error('act: 调用对话行为分类器失败， 原因：' + msg)

    def predict(self, sentence: str, acts: list = None):
        if not acts:
            return self._predict_all(sentence)
        else:
            params = {
                'query': sentence,
                'types': acts
            }
            try:
                response = requests.post(url=self.act_predict_url, json=params)
                if self._response_failed(response):
                    logging.error("act: 调用对话行为分类器失败: 接口异常。" + str(response.json()) if response is not None else '')
                else:
                    acts = response.json()['reply']['labels']
                    return [
                        {
                            'act': act['label'], 'score': act['confidence']}
                        for act in acts
                    ]
            except Exception:
                msg = "Exception When POST '{}' : {}".format(self.act_predict_url, traceback.format_exc())
                logging.error('act: 调用对话行为分类器失败， 原因：' + msg)

    def get_acts(self):
        try:
            response = requests.get(url=self.sys_act_get_url)
            if self._response_failed(response):
                logging.error("act: 调用对话行为分类器失败: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                acts = response.json()['sys_dialog_acts']
                return [
                    {
                        'name': act['cn_name'],
                        'id': act['name'],
                        'explanation': act['explaination'],
                        'samples': act['typical_expression']
                    }
                    for act in acts
                ]
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.sys_act_get_url, traceback.format_exc())
            logging.error('act: 调用对话行为分类器失败， 原因：' + msg)



import logging
import sys

log = logging.getLogger(__name__)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)

formatter = logging.Formatter(
    '%(asctime)s %(levelname)s --- [bf-engine-sdk]%(message)s'
)
handler.setFormatter(formatter)

log.addHandler(handler)
log.setLevel(logging.INFO)



import os
from . import util
from .config import Config
from .core.bot import Bot
from .utils.local_cache import LocalCache
import urllib3

urllib3.disable_warnings()

"""
default init all modules
"""

def init(app_id: str = None, local=False, url=None, init_all_modules=True) -> Bot:
    """
    创建bot，并返回实例
    """
    tmp_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
    if not app_id:
        # app_id = util.random_uui
        cache = LocalCache(tmp_dir + "/var.bot")
        if cache:
            app_id = cache["app_id"]
        if not app_id:
            app_id = util.random_uuid()
            cache["app_id"] = app_id
    if local:
        Config.base_url = 'http://172.17.0.1'
    elif url:
        Config.base_url = url
    else:
        Config.base_url = Config.remote_url

    bot = Bot(app_id, init_all_modules)
    return bot





import random
import string
import uuid


def random_str(length):
    salt = ''.join(random.sample(string.ascii_letters + string.digits, length))

    return salt


def random_uuid():
    return str(uuid.uuid4()).replace('-', '')



import logging
import traceback

import requests

from ..config import Config
from ..entity.answer import Answer
from ..entity.exception import BfEngineException, ErrorCode
from .base import CallerBase


class DialogueManagerCaller(CallerBase):

    def __init__(self, app_id):
        super().__init__(app_id)
        self.app_id = app_id
        self.load_config_url = "{}/validate".format(Config.base_url)
        self.call_url = "{}/v1/ccs/openapi".format(Config.base_url)

    def load_config(self, data: dict) -> bool:
        """
        加载配置
        :param data:
        :return:
        """
        ret = True
        try:
            response = requests.post(self.load_config_url, json=data)
            if self._response_failed(response) or response.json().get('code') != 'Success':
                msg = "kg: 加载对话配置失败：原因: 接口返回异常。" + str(response.json()) if response is not None else ''
                raise BfEngineException(ErrorCode.dm_load_error, msg)
        except BfEngineException as bfe:
            raise bfe
        except Exception:
            msg = "kg: 加载对话配置失败， 原因: Exception When POST '{}' : {}".format(self.load_config_url, traceback.format_exc())
            raise BfEngineException(ErrorCode.dm_load_error, msg)
        return ret

    def call(self, msg):
        """
       预测出话
       :param msg: 用户问
       :return: 机器人回答
       """
        answer = Answer(text='', score=0)
        param = {
            "text": msg,
            "extendData": {
                "online": False
            },
        }
        header = {
            'appId': self.app_id,
            'sessionId': Config.session_id,
            'userId': Config.user_id,
            'Content-Type': 'application/json'
        }
        try:
            response = requests.post(url=self.call_url, headers=header, json=param)
            if self._response_failed(response):
                logging.error("dm: 预测失败：原因: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                if response_json.get('data'):
                    datas = response_json['data']
                    answer_text = ";".join([data.get('value') or '' for data in datas])
                    score = 100
                    if response_json.get('info') and response_json['info']['textScore']:
                        score = float(response_json['info']['textScore'])
                    answer = Answer(text=answer_text, score=score)
                else:
                    pass
                    # logging.warning('dm: 预测为空 [{}]'.format(msg))
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.call_url, traceback.format_exc())
            logging.error('dm: 对话失败， 原因: ' + msg)
        finally:
            if answer.text == '无候选答案' or not answer.text:
                answer.text = Config.backfill_speech
                answer.score = 60
            return answer



import time

from .module import Module
from ..caller.config_manager_caller import ConfigManagerCaller


class ConfigManager(Module):
    def __init__(self, app_id, set, init):
        super().__init__(app_id, 'config', set)
        self.app_id = app_id
        self.caller = ConfigManagerCaller(self.app_id)
        if init:
            self.init()

    def init(self):
        self.caller.init_bot_config()

    def get_backfill_answers(self) -> list:
        """
        获取未知回复列表
        :return:
        """
        answers = self.caller.get_backfill()
        return [answer['content'] for answer in answers]

    def set_backfill_answer(self, answers: list) -> bool:
        result = self.caller.set_backfill(answers)
        # todo delete
        time.sleep(1.5)
        #

        return result



from .module import Module
from ..caller.act_caller import DialogueActCaller


class ActAnswer:
    def __init__(self, name, code, score):
        """
        :param name: 命中对话行为分类器的名称
        :param code: 命中对话行为分类器的code
        :param score: 对话行为分类器的出分
        """
        self.name = name
        self.code = code
        self.score = score


class Act:
    def __init__(self, name, code, explanation, sample):
        """
        :param name: 对话行为分类器名称
        :param code: 对话行为分类器调用code
        :param explanation: 对话行为分类器作用释义
        :param sample: 调用示例
        """
        self.name = name
        self.code = code
        self.explanation = explanation
        self.sample = sample


class DialogueAct(Module):

    def __init__(self, app_id, set, init):
        super().__init__(app_id, 'ner', set)
        self.app_id = app_id
        self.caller = DialogueActCaller(app_id)
        if init:
            self.init()

    def init(self):
        act_code_map = {}
        acts = self.get_acts()
        for act in acts:
            act_code_map[act.code] = act
        self.act_code_map = act_code_map

    def predict(self, sentence: str, acts: list = None):
        """
        行为分类器预测接口
        :param sentence: 用户输入
        :param acts: 调用分类器的code列表， 为空或者不传则默认调用所有分类器
        :return: list(ActAnswer)
        """
        results = self.caller.predict(sentence, acts)
        answers = []
        if results:
            for result in results:
                hit_act = result['act']
                answers.append(
                    ActAnswer(hit_act, self.act_code_map.get(hit_act).name if hit_act in self.act_code_map else '',
                              result['score']))
        return answers

    def get_acts(self):
        """
        获取系统内存在的对话行为分类器列表
        :return: list(Act)
        """
        acts = self.caller.get_acts() or []
        return [
            Act(act['name'], act['id'], act['explanation'], act['samples'])
            for act in acts
        ]



import time
import traceback
import os

from tqdm import tqdm

from .module import Module
from ..caller.qa_caller import QACaller
from ..entity.answer import Answer
from ..entity.enums import QuestionType, QuestionMode,QuestionField
from ..entity.exception import BfEngineException, ErrorCode
from ..logger import log


class QuestionAnswering(Module):
    """
    问答
    """
    tmp_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
    faq_sq_empty_path = tmp_dir + "/data/问答上传模板.xlsx"
    faq_lq_empty_path = tmp_dir + "/data/语料上传模板.xlsx"

    def __init__(self, app_id, set, init):
        super().__init__(app_id, 'qa',set)
        self.caller = QACaller(app_id)
        if init:
            self.init()

    def train(self, data: str = None, question_path: str = None, corpus_path: str = None,append: bool = False,show_progress:bool=True):
        """
        :param data:
        :param question_path: 标准问路径
        :param corpus_path: 语料路径
        :param append: 是否追加，True
        :return: 训练是否成功
        """
        try:
            log.info('qa: prepare train')
            if data:
                # 清空FAQ
                if not append:
                    self._upload_question(self.faq_sq_empty_path, is_log=False)
                    self._upload_corpus(self.faq_lq_empty_path, is_log=False)
                self._upload_json(data)
            else:
                self._upload_question(question_path,append)
                if corpus_path:
                    self._upload_corpus(corpus_path,append)
            return self._train(show_progress)
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error('unknown exception')
            log.error(e)
            traceback.print_exc()
            return False
        return True
    def train_status(self,train_id:str=None):
        if not train_id:
            raise BfEngineException(code=-1, msg="train_id is null")
        return self.caller.train_status(train_id)
    def publish(self):
        """
        :发布
        :return: 机器人回答
        """
        self._pulish()
    def export(self,sq_path: str = None, lq_path: str = None):
        """
        导出问答和语料文件
        :param sq_path 问答文件路径
        :param lq_path 语料文件路径
        """
        self._export(sq_path,lq_path)

    def query(self, text: str,online: bool=False) -> Answer:
        """
        :param text: 用户问
        :return: 机器人回答
        """
        return self.caller.call_module('faq', text,online)
    def qa_tag_list(self):
        """
        获取标签列表
        """
        return self.caller.qa_tag_list()
    def qa_tag_add(self,name:str=None):
        """
        添加标签
        :param name 标签名称
        """
        return self.caller.qa_tag_add(name)
    def qa_list(self) -> str:
        """
        : 查询标准问列表
        :return: 标准问列表
        """
        return self.caller.qa_list(QuestionField.SQ)
    def question_list(self) -> str:
        """
        : 查询标准问列表
        :return: 标准问列表
        """
        return self.caller.sq_list(QuestionField.ALL)

    def corpus_list(self,question_id) -> str:
        """
        : 查询标准问下语料列表
        :return: 标准问列表
        """
        return self.caller.lq_list(question_id)
    def _train(self,show_progress:bool=True):
        log.info('qa: start training')
        train_id = self.caller.train()
        if not show_progress:
            return train_id
        progress = 0

        with tqdm(total=100) as bar:
            bar.set_description("qa training...")
            while progress < 100:
                old_progress = progress
                progress = self.caller.train_status(train_id)
                bar.update(progress - old_progress)
                time.sleep(1)
            bar.set_description("qa train finished")
        time.sleep(5)
    def _pulish(self):
        log.info('qa: start release')
        publish_id = self.caller.publish()
        progress = 0

        with tqdm(total=100) as bar:
            bar.set_description("qa release...")
            while progress < 100:
                old_progress = progress
                progress = self.caller.publish_status(publish_id)
                bar.update(progress - old_progress)
                time.sleep(1)
            bar.set_description("qa release finished")
        time.sleep(5)
    def _upload_json(self, content):
        """
        获取从路径中json
        """
        try:
            data = content["data"]
            for item in data:
                data_id = self.caller.upload_json_sq(self.set.tag,item)
                self.caller.upload_json_lq(data_id, item)
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            return None

    def _upload_corpus(self, corpus_path, is_log: bool = True,append: bool = False):
        """
        上传qa语料
        :param corpus_path 语料路径
        :param append True:增量,false:全量
        """
        data_id = self.caller.upload(QuestionType.LQ, QuestionMode.INCRE if append else QuestionMode.FULL, corpus_path)

        progress = 0
        while progress < 100:
            progress = self.caller.upload_status(data_id, is_log=is_log)
            time.sleep(1)

    def _upload_question(self, question_path, is_log=True,append: bool = False):
        if not question_path:
            raise BfEngineException(ErrorCode.argument_missing_error, 'missing question path')

        data_id = self.caller.upload(QuestionType.SQ_ANS, QuestionMode.INCRE if append else QuestionMode.FULL, question_path)

        progress = 0
        while progress < 100:
            progress = self.caller.upload_status(data_id, is_log=is_log)
            time.sleep(1)
    def _export(self,sq_path: str = None, lq_path: str = None):
        """
        导出问答和语料文件
        :param sq_path 问答文件路径
        :param lq_path 语料文件路径
        """
        try:
            if sq_path:
                data_path = self.caller.download_launch(QuestionType.SQ_ANS)
                self.caller.download(sq_path,data_path)
            if lq_path:
                data_path = self.caller.download_launch(QuestionType.LQ)
                self.caller.download(lq_path, data_path)
        except BfEngineException as bfe:
            log.error(bfe)
        except Exception as e:
            log.error(e)

    def update_corpus(self, question_id, corpus_id, corpus):
        """
        更新语料
        :return:
        """
        return self.caller.update_corpus(question_id, corpus_id, corpus)

    def add_corpus(self, question_id, corpus):
        """
        增加语料
        :return:
        """
        lq_id = self.caller.add_corpus(question_id, corpus)
        return lq_id

    def delete_corpus(self, corpus_id):
        """
        删除语料
        :return:
        """
        return self.caller.delete_corpus(corpus_id)

    def add_question(self, question_data: dict):
        """
        添加问题
        :param question_data:
        :return:
        """
        question_id = self.caller.upload_json_sq(data=question_data)
        self.caller.upload_json_lq(question_id, question_data)
        return question_id


    def delete_question(self, question_id: int):
        """
        删除问题
        :param question_id:
        :return:
        """
        return self.caller.delete_question(question_id)

    def update_question(self, question_data: dict):
        """
        更新问题
        :param question_data:
        :return:
        """
        return self.caller.update_question(question_data)

    def update_answer(selfs, answer_data: dict):
        """
        更新答案
        """
        return  selfs.caller.update_answer(answer_data)

    def query_category_list(self) -> str:
        return self.caller.query_category_list()

    def add_category(self,parent_id,name:str):
        return self.caller.add_category(parent_id,name)

    def update_category(self,id,name:str):
        return self.caller.update_category(id,name)

    def delete_category(self,id):
        return self.caller.delete_category(id)

    def get_similar_qa_list(self,user_inputs:str,max_topn,category_ids,online: bool=True) -> str:
        return self.caller.get_similar_qa_list(user_inputs,max_topn,category_ids,online)

    def get_similar_recomend_qa_list(self,user_inputs:str,type:str) -> str:
        """
        获取相似问/推荐问
        :param user_inputs 用户问
        :param type similar:相似问，recommend:推荐问
        """
        return self.caller.get_similar_recomend_qa_list(user_inputs,type)




import json
import copy
import os
import traceback
import qrcode
# import qrcode_terminal
import uuid
from .. import util

from ..caller.dialogue_manager_caller import DialogueManagerCaller
from ..config import Config
from ..core.module import Module
from ..entity.exception import BfEngineException
from ..logger import log
from .task_engine import TaskEngine
module_name_map = {
    "qa": "faq",
    "kg": "domain_kg",
    "te": "task_engine",
    "backfill": "backfill"
}
module_template = {
    "ccs_id": "",
    "name": "",
    "url": "",
    "app_id": "",
    "modules": [],
    "request_header": "",
    "request_body": "",
    "active": True,
    "timeout": 3500,
    "scenario_id": "",
    "subscribeKeys": {},
    "publishKeys": {}
}
arg_template = {
    "ccs_id": "",
    "name": "",
    "path": "",
    "value": ""
}
model_template = {
    "ccs_id": "",
    "name": "",
    "url": "",
    "method": "post",
    "category": "external",
    "request_body": "",
    "request_header": "",
    "app_id": ""
}

phase_template = {
    "phase_id": "",
    "action_rules": [],
    "group": [],
    "dispatcher": {"applyCurTask": False, "applyPreTask": False, "dispatchers": []},
    "score_rules": [],
    "respond_rules": [],
    "priority_rules": [],
    "thresholds": [],
    "bots": [],
    "models": [],
    "actions": []
}

phase_group_template = {"name": "", "flowMode": "ALL", "bots": []}
phase_dispatcher_template = {"exclusive": False, "filter": {"op": "and", "operands": []}, "groups": []}

dispatcher_operands_template = {
    "op": "",
    "operand1": {
        "op": "select",
        "path": ""
    },
    "operand2": ""
}

threshold_template = {
    "condition": {},
    "threshold": 60,
    "bot": ""
}
config_template = {
    "ccs_id": "",
    "confid": 2020,
    "name": "dialogue_manager_config",
    "status": "active",
    "ranker": {"respond_rules": [], "priority_rules": [], "score_rules": [], "bots": []},
    "chat_config": {"chat_config": {"dispatchers": []}, "bots": [], "models": []},
    "workflow": []
}
init_models = [
    {
        "uid": "intent_model",
        "ccs_id": "",
        "name": "意图",
        "url": "http://172.17.0.1:11035/predict",
        "method": "post",
        "category": "external",
        "request_body": {
            "SessionId": "$.request.sessionId",
            "Text": "$.request.body.text",
            "Robot": "$.request.appId",
            "IsRelease": True,
            "UserId": "$.request.userId"
        },
        "request_header": {
            "X-locale": "zh-cn",
            "app_id": "",
            "user_id": "bf-engine-sdk",
            "Authorization": "Bearer EMOTIBOTDEBUGGER",
            "Accept": "application/json,text/plain, */*"
        },
        "app_id": ""
    }
]
init_args = [
    {
        "uid": "intent_arg",
        "ccs_id": "",
        "name": "$intent",
        "path": "$.model.意图.data[0].matchQuestion",
        "value": ""
    }
]


class ConfigGenerator:
    @staticmethod
    def get_init_models(app_id, url=None):
        models = copy.deepcopy(init_models)
        for model in models:
            request_header = model.get('request_header')
            request_body = model.get('request_body')
            model['ccs_id'] = app_id
            model['app_id'] = app_id
            if url is not None:
                model['url'] = url
            if type(request_header) == dict:
                model['request_header']['app_id'] = app_id
                model['request_header'] = json.dumps(request_header)
            if type(request_body) == dict:
                model['request_body']['Robot'] = app_id
                model['request_body'] = json.dumps(request_body)
        return models

    @staticmethod
    def get_module(module: str, app_id: str, url: str, request_header: dict = {}, request_body: dict = {}):
        module = module_name_map[module]
        name = util.random_uuid()
        module_object = copy.deepcopy(module_template)
        module_object['ccs_id'] = app_id
        module_object['name'] = name
        module_object['url'] = url
        module_object['app_id'] = app_id
        module_object['modules'].append(module)
        module_object['request_header'] = json.dumps(request_header)
        module_object['request_body'] = json.dumps(request_body)
        return module_object

    @staticmethod
    def get_model(app_id: str, name: str, url: str, request_header: dict, request_body: dict, method='post',
                  category='external'):
        assert name and url
        model = copy.deepcopy(model_template)
        model['ccs_id'] = app_id
        model['app_id'] = app_id
        model['name'] = name
        model['method'] = method
        model['url'] = url
        model['request_header'] = json.dumps(request_header, indent=2)
        model['request_body'] = json.dumps(request_body, indent=2)
        model['category'] = category
        return model

    @staticmethod
    def get_arg(app_id: str, name: str, json_path: str):
        arg = copy.deepcopy(arg_template)
        arg['name'] = name
        arg['ccs_id'] = app_id
        arg['path'] = json_path
        return arg

    @staticmethod
    def get_config(app_id: str):
        config = copy.deepcopy(config_template)
        config['ccs_id'] = app_id
        return config

    @staticmethod
    def get_phase():
        return copy.deepcopy(phase_template)

    @staticmethod
    def get_phase_group():
        return copy.deepcopy(phase_group_template)

    @staticmethod
    def get_phase_dispatcher():
        return copy.deepcopy(phase_dispatcher_template)

    @staticmethod
    def get_dispatcher_operand(arg_value, path, condition='eq'):
        operand = copy.deepcopy(dispatcher_operands_template)
        operand['op'] = condition
        operand['operand1']['path'] = path
        operand['operand2'] = arg_value
        return operand

    @staticmethod
    def get_threshold(module_name: str, threshold_score: int):
        threshold = copy.deepcopy(threshold_template)
        threshold['bot'] = module_name
        threshold['threshold'] = threshold_score
        return threshold

    @staticmethod
    def get_score_rule(module_name: str, score: int):
        return {
            "adjustment": str(score),
            "condition": {},
            "unit": module_name,
            "id": util.random_uuid()
        }



class DialogueManager(Module):
    """
    对话管理
    """

    def __init__(self, app_id, set, init):
        super().__init__(app_id, 'dm', set)
        self.app_id = app_id
        self.module_call_url = '{}/{}'.format(Config.base_url, Config.robot_url)
        self.chat_url = '{}/{}?appId={}'.format(Config.base_url, Config.chat_url,app_id)
        self.caller = DialogueManagerCaller(app_id)
        tmp_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        self.default_dm_config_path = tmp_dir + "/data/default_dm.json"
        self.task_engine_service = TaskEngine(app_id, set, init)
        if init:
            self.init()

    def init(self):
        self.load()

    """
    加载对话管理配置，不传配置数据则加载默认配置
    
    """

    def load(self, data: dict = None, path: str = None):
        try:

            if not data and path:
                data = json.load(open(path))
            elif not data:
                data = json.load(open(self.default_dm_config_path))
            config = self._config_generate(data)
            return self.caller.load_config(config)
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error("dm: 加载对话配置异常")
            log.error(e)
            traceback.print_exc()
            return False

    def load_by_path(self, config_path: str = None):
        simple_phases = json.load(open(config_path))
        if not config_path:
            simple_phases = json.load(open(self.default_dm_config_path))
        return self.load(simple_phases)


    """
    根据传入参数生成出话配置
    """

    def _config_preprocess(self, phases):
        modules = []
        for phase in phases:
            for module_name, module_strategy in phase.items():
                assert module_name in module_name_map, 'module just accept in [te, kg, qa]'
                module_object = ConfigGenerator.get_module(module_name, self.app_id, self.module_call_url)
                if type(module_strategy) == dict:
                    module_strategy['module_name'] = module_object['name']
                    if module_name == 'te' and 'name' in module_strategy:
                        scenario_name = module_strategy['name']
                        if scenario_name:
                            scenario_id = self.task_engine_service.get_taskengine_scenario_id(scenario_name)
                            assert scenario_id, 'te scenario name [{}] not found'.format(scenario_name)
                            module_object['scenario_id'] = scenario_id
                elif type(module_strategy) == int:
                    phase[module_name] = {
                        'threshold': int(module_strategy),
                        'module_name': module_object['name']
                    }
                if 'threshold' not in phase[module_name]:
                    phase[module_name]['threshold'] = 60
                modules.append(module_object)
        return modules

    def _config_generate(self, data):
        workflow = []
        simple_phases = data
        args = copy.deepcopy(init_args)
        arg_map = {
            arg['name']: arg
            for arg in args
        }
        modules = self._config_preprocess(simple_phases)
        config = ConfigGenerator.get_config(self.app_id)
        config['workflow'] = workflow

        # 现在只有预置模型服务
        models = ConfigGenerator.get_init_models(self.app_id)
        # 增加backfill模块
        backfill_module_object = ConfigGenerator.get_module('backfill', self.app_id, self.module_call_url)
        modules.append(backfill_module_object)

        # current task for te
        init_phase = ConfigGenerator.get_phase()
        init_phase['phase_id'] = util.random_uuid()
        init_phase['dispatcher']['applyCurTask'] = True
        init_phase['bots'] = modules
        for module in modules:
            if "task_engine" in module['modules']:
                threshold = ConfigGenerator.get_threshold(module['name'], 90)
                init_phase['thresholds'].append(threshold)
        workflow.append(init_phase)

        for simple_phase in simple_phases:
            phase = ConfigGenerator.get_phase()
            for phase_module in simple_phase.keys():
                module_condition = simple_phase[phase_module]
                threshold_score = module_condition['threshold']
                group = ConfigGenerator.get_phase_group()
                group_name = util.random_uuid()
                group['name'] = group_name
                group['bots'].append(module_condition['module_name'])
                phase['group'].append(group)

                dispatcher = ConfigGenerator.get_phase_dispatcher()
                dispatcher['groups'].append(group_name)
                phase['dispatcher']['dispatchers'].append(dispatcher)
                if 'condition' in module_condition:
                    condition = module_condition['condition']
                    for arg_key in condition:
                        if arg_key in arg_map:
                            path = arg_map[arg_key]['path']
                            operand = ConfigGenerator.get_dispatcher_operand(condition[arg_key], path)
                            dispatcher['filter']['operands'].append(operand)
                        else:
                            log.warning("arg {} not found".format(arg_key))

                threshold = ConfigGenerator.get_threshold(module_condition['module_name'], threshold_score)
                phase['thresholds'].append(threshold)

            phase['phase_id'] = util.random_uuid()

            phase['bots'] = modules
            phase['models'] = models
            workflow.append(phase)
        # 附加 当前多轮 追问出话策略
        task_engine_rewind_phase = ConfigGenerator.get_phase()
        task_engine_rewind_phase['phase_id'] = util.random_uuid()
        task_engine_rewind_phase['dispatcher']['applyCurTask'] = True
        task_engine_rewind_phase['bots'] = modules
        for module in modules:
            if "task_engine" in module['modules']:
                threshold = ConfigGenerator.get_threshold(module['name'], 75)
                task_engine_rewind_phase['thresholds'].append(threshold)
        workflow.append(task_engine_rewind_phase)

        # 附加backfill出话模块
        backfill_phase = ConfigGenerator.get_phase()
        backfill_phase['phase_id'] = util.random_uuid()
        backfill_phase['bots'] = modules
        group = ConfigGenerator.get_phase_group()
        group_name = util.random_uuid()
        group['name'] = group_name
        group['bots'].append(backfill_module_object['name'])
        backfill_phase['group'].append(group)
        backfill_dispatcher = ConfigGenerator.get_phase_dispatcher()
        backfill_dispatcher['groups'].append(group_name)
        backfill_phase['dispatcher']['dispatchers'].append(backfill_dispatcher)
        backfill_phase['score_rules'].append(ConfigGenerator.get_score_rule(backfill_module_object['name'], 90))
        workflow.append(backfill_phase)

        return config
    def qrcode(self,filePath:str=None):
        if filePath:
            qr = qrcode.QRCode(
                version=None,
                error_correction=qrcode.constants.ERROR_CORRECT_L,
                box_size=10,
                border=2,
            )
            qr.add_data(self.chat_url)
            qr.make(fit=True)
            img = qr.make_image()
            img.save(filePath)
        else:
            return
            # qrcode_terminal.draw(self.chat_url)
    def query(self, msg):
        try:
            answer = self.caller.call(msg)
            return answer
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception:
            log.error("dm: 出话异常")
            traceback.print_exc()
            return False

    """
    检查传入配置
    """

    def _check_config(self):
        # todo check model arg module repeat name
        pass



from ..logger import log


class TestSet:
    def __init__(self, cases):
        self.cases = []
        for actual, expect in cases:
            self.cases.append(TestCase(actual, expect))

    def __iter__(self):
        return self.cases.__iter__()

    def __len__(self):
        return len(self.cases)


class TestCase:
    def __init__(self, text, expect):
        self.text = text
        self.expect = expect
        self.actual = None
        self.result = None
        self.done = False

    def run(self, module):
        self.actual = module.query(self.text).text
        if self.actual == self.expect:
            self.result = True
        else:
            self.result = False
        self.done = True

    def __str__(self):
        if self.done:
            if self.result:
                return 'passed, case: {}, expect: {}, actual: {}'.format(self.text, self.expect, self.actual)
            return 'failed, case: {}, expect: {}, actual: {}'.format(self.text, self.expect, self.actual)
        return 'case: {}, expect: {}'.format(self.text, self.expect)


class TestResult:
    def __init__(self, testset):
        self.testset = testset
        self.total = len(testset)
        self.wrong_cases = list(filter(lambda c: not c.result, self.testset))
        self.correct = self.total - len(self.wrong_cases)

    def show(self):
        log.info('report start')
        log.info('total: {}'.format(self.total))
        log.info('correct: {}'.format(self.correct))
        log.info('accuracy: {}'.format(self.correct / self.total))
        for case in self.wrong_cases:
            log.info(case)
        log.info('report end')
        log.info('-' * 20)


class TestRunner:
    def __init__(self, testset, module):
        self.testset = testset
        self.module = module
        self.result = None

    def run(self):
        for case in self.testset:
            case.run(self.module)

        return TestResult(self.testset)



import json
import time
import traceback
import os
from selenium import webdriver
from selenium.webdriver.remote.command import Command
from webdriver_manager.chrome import ChromeDriverManager

import requests
from requests import Session, Request

from .module import Module
from ..config import Config
from ..entity.answer import Answer
from ..logger import log


class TaskEngine(Module):
    """
    任务引擎
    """

    def __init__(self, app_id, set, init):
        super().__init__(app_id, 'te', set)
        self.app_id = app_id
        self.sess = Session()
        self.api_server = Config.base_url
        tmp_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        self.chrome_path_bin = tmp_dir + "/data/bin"
        # 获取scenario列表
        self.query_scenarios_url = '{}/php/api/ApiKey/task_engine_app.php'.format(
            self.api_server)
        # 获取scenario详细信息
        self.query_scenario_info_url = '{}/php/api/ApiKey/task_engine_scenario.php'.format(
            self.api_server)
        self.editor_url = '{}/?appid={}#/task-engine-scenario-v2'.format(
            self.api_server, self.app_id)
        # 获取scenario详细信息
        self.query_url = '{}/task_engine/ET'.format(self.api_server)
        self.headers = {
            'X-Appid': self.app_id,
            'X-Userid': Config.user_id
        }
        if init:
            self.init()

    def load(self, path):
        """
        加载scenario 的json
        """
        scenario_json = self._get_scenariojson_from_path(path)
        scenario_id, scenario_name = self._add_scenarios(scenario_json)
        if scenario_id is None:
            return False
        else:
            if self.enable_scenario(scenario_id):
                if self._public_scenario(scenario_id):
                    time.sleep(10)
                    log.info("te: load scenario【" + scenario_name + "】success")
                    return True
        log.error("te: load scenario【" + scenario_name + "】failed")
        return False

    def editor(self,path:str=None,syn:bool=True):
        """

        """
        if path:
            self.load(path)
        driver = webdriver.Chrome(ChromeDriverManager().install())
        if driver:
            driver.get(self.editor_url)
            # while self.driver_status(driver):
            #     time.sleep(5)
            # func()
        log.info("正在为您打开页面:" + self.editor_url)
        if syn:
            log.info("任务引擎编辑: 编辑好后请按[return]键进行下一步操作")
            while True:
                file_content = input()
                if file_content == '':
                    break

    def driver_status(self,driver:webdriver)->bool:
        try:
            driver.execute(Command.STATUS)
            return True
        except Exception:
            return False
    def query(self, txt):
        """
        获取对话信息
        """
        try:
            task_engine_json = self._get_taskengine_json(txt, self.app_id)
            headers = {'Content-Type': 'application/json;charset=UTF-8'}
            resp = requests.post(self.query_url, json=task_engine_json, headers=headers).json()
            if "TTSText" not in resp or resp["TTSText"] is None:
                Answer(text=Config.backfill_speech, score=60)
            else:
                return Answer(text=resp["TTSText"], score=100)
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_url, traceback.format_exc())
            log.error('te: 对话失败， 原因: ' + msg)
        return Answer(text=Config.backfill_speech, score=60)

    def _query_scenarios(self):
        """
        查询scenario列表
        """
        try:
            params = {"appid": self.app_id, "type": 0}
            req = Request('GET', self.query_scenarios_url, headers=self.headers, params=params)
            resp = self.sess.send(req.prepare(), verify=False, timeout=2).json()
            return resp['msg']
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_scenarios_url, traceback.format_exc())
            log.error('te: 查询scenario列表失败， 原因: ' + msg)
        return []

    def _query_scenario_info(self, scenario_id):
        """
         查询scenario详细信息
         """
        try:
            params = {"scenarioid": scenario_id}
            req = Request('GET', self.query_scenario_info_url, headers=self.headers, params=params)
            resp = self.sess.send(req.prepare(), verify=False, timeout=2).json()
            return resp
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_scenario_info_url, traceback.format_exc())
            log.error('te: 查询scenario【' + scenario_id + '】信息， 原因: ' + msg)
        return []

    def _add_scenarios(self, scenario_json):
        """
        导入scenario
        """
        try:
            metadata = None
            scenario_id, scenario_name = self._get_scenario_info(scenario_json)
            scenario_lists = self._query_scenarios()
            for scenario in scenario_lists:
                if 'scenarioID' in scenario and scenario['scenarioID'] == scenario_id:
                    metadata = self._create_scenario(scenario_name)
                    if metadata is None:
                        return None, None
                    break
            content = scenario_json['taskScenario']
            layout = scenario_json['taskLayouts']
            params = {
                'content': json.dumps(content),
                'scenarioid': scenario_id,
                "appid": self.app_id,
                'layout': json.dumps(layout),
                'method': 'PUT'
            }
            if metadata is not None:
                content['metadata'] = metadata
                scenario_id = metadata['scenario_id']
                scenario_name = metadata['scenario_name']
            else:
                params['upload'] = 1
                params['type'] = 0

            resp = requests.post(url=self.query_scenario_info_url, data=params, headers=self.headers)
            if resp is None or resp.status_code != 200:
                return None, None
            return scenario_id, scenario_name
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_scenario_info_url, traceback.format_exc())
            log.error('te: 导入scenario失败， 原因: ' + msg)

        return None, None

    def _get_scenario_info(self, scenario_json):
        """
        获取scenario配置信息
        """
        try:
            return scenario_json["taskScenario"]["metadata"]["scenario_id"], scenario_json["taskScenario"]["metadata"][
                "scenario_name"]
        except Exception:
            pass
        return None, None

    def enable_scenario(self, scenario_id):
        """
        启动scenario
        """
        try:
            params = {'scenarioid': scenario_id, "appid": self.app_id, 'enable': 'true'}
            req = Request('POST', self.query_scenarios_url, headers=self.headers, params=params)
            resp = self.sess.send(req.prepare(), verify=False, timeout=2)
            if resp is None or resp.status_code != 200:
                log.error("te: 启用scenario【" + scenario_id + "】失败。")
                return False
            return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_scenarios_url, traceback.format_exc())
            log.error('te: 启用scenario【' + scenario_id + '】失败， 原因: ' + msg)
        return False

    def _disable_scenario(self, scenario_id):
        """
        停用scenario
        """
        try:
            params = {'scenarioid': scenario_id, "appid": self.app_id, 'enable': 'false'}
            req = Request('POST', self.query_scenarios_url, headers=self.headers, params=params)
            resp = self.sess.send(req.prepare(), verify=False, timeout=2)
            if resp is None or resp.status_code != 200:
                log.error("te: 停用scenario【" + scenario_id + "】失败。")
                return False
            return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_scenarios_url, traceback.format_exc())
            log.error('te: 停用scenario【' + scenario_id + '】失败， 原因: ' + msg)
        return False

    def _public_scenario(self, scenario_id):
        """
        发布scenario
        """
        try:
            params = {'scenarioid': scenario_id, "appid": self.app_id, 'publish': 1}
            req = Request('PUT', self.query_scenario_info_url, headers=self.headers, params=params)
            resp = self.sess.send(req.prepare(), verify=False, timeout=2)
            if resp is None or resp.status_code != 200:
                log.error("te: 发布scenario【" + scenario_id + "】失败。")
                return False
            return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_scenario_info_url, traceback.format_exc())
            log.error('te: 发布scenario【' + scenario_id + '】失败， 原因: ' + msg)
            return False
        finally:
            time.sleep(1)

    def _create_scenario(self, scenario_name):
        """
        新建scenario
        """
        try:
            params = {"scenarioName": scenario_name, "template": '', "appid": self.app_id, "type": 0}
            req = Request('POST', self.query_scenario_info_url, headers=self.headers, params=params)
            resp = self.sess.send(req.prepare(), verify=False, timeout=2).json()
            return resp['template']['metadata']
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_scenario_info_url, traceback.format_exc())
            log.error('te: 新建scenario【' + scenario_name + '】失败， 原因: ' + msg)
        return None

    def _del_scenarios(self, scenario_id):
        """
        删除scenario
        """
        try:
            params = {"scenarioid": scenario_id, "delete": 1}
            req = Request('PUT', self.query_scenario_info_url, headers=self.headers, params=params)
            resp = self.sess.send(req.prepare(), verify=False, timeout=2)
            if resp is None or resp.status_code != 200:
                log.error("te: 删除scenario【" + scenario_id + "】失败。")
                return False
            return True
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.query_scenario_info_url, traceback.format_exc())
            log.error('te: 删除scenario【' + scenario_id + '】失败， 原因: ' + msg)

        return False

    def _get_scenariojson_from_path(self, path):
        """
        获取从路径中json
        """
        try:
            with open(path, encoding='utf-8-sig', errors='ignore') as f:
                data = json.load(f, strict=False)
                if type(data) is list:
                    return data[0]
                elif type(data) is dict:
                    return data
        except Exception:
            return None

    def _get_taskengine_json(self, txt, app_id):
        """
        生成scenario对话json
        """
        te_json = {}
        text_conversion = {'text_simplified': txt, 'text_traditional': txt}
        te_json["TextConversion"] = text_conversion
        te_json["Text"] = txt
        te_json["UserID"] = Config.user_id
        te_json["ExtendData"] = {}
        te_json["SessionID"] = Config.session_id
        te_json["UniqueID"] = Config.user_id + Config.session_id
        te_json["AppID"] = app_id
        config = {"webapi_timeout": 1000}
        te_json["Config"] = config
        return te_json

    def get_taskengine_scenario_id(self, scenario_name):
        try:
            scenario_list = self._query_scenarios() or []
            scenario = list(filter(lambda x: x['scenarioName'] == scenario_name, scenario_list))
            if len(scenario) > 0:
                return scenario[0]['scenarioID']
        except Exception:
            return None



from ..core.test import TestRunner
from ..entity.answer import Answer
from .set import Set


class Module:
    """
    出话模块
    """
    def __init__(self, app_id, name, set):
        self.app_id = app_id
        self.name = name
        self.set = set

    def query(self, text: str, online: bool) -> Answer:
        """
        :param text: 用户问
        :param online: True 线上出话 False 线下出话
        :return: 机器人回答
        """
        raise NotImplemented

    def test(self, dataset):
        """
        :param dataset:
        :return: 测试结果
        """
        runner = TestRunner(dataset, self)
        result = runner.run()
        return result

    def init(self):
        pass



import shelve
from contextlib import closing

class LocalCache:
    def __init__(self, cache='var.pkl'):
        self.cache = cache

    def __setitem__(self, key, value):
        """
        key: 变量名
        value: 变量值
        cache: 缓存名
        """
        with closing(shelve.open(self.cache, 'c')) as shelf:
            shelf[key] = value

    def __getitem__(self, key):
        """
        key : 变量名
        return：变量值
        """
        with closing(shelve.open(self.cache, 'c')) as shelf:
            return shelf.get(key)

    def remove(self, key):
        """
    	key: 变量名

    	如果 变量存在则删除, 如果不存在，会抛出一个异常，由调用方去处理
    	"""
        with closing(shelve.open(self.cache, 'rc')) as shelf:
            del shelf[key]
            return True


import logging
import traceback

import requests

from .base import CallerBase



class NERParserCaller(CallerBase):
    """
    通用解析器API
    """
    def __init__(self, app_id):
        super().__init__(app_id)
        self.app_id = app_id
        self.module = 'ner'
        self.tde_user_register_url = self.base_url + '/nerFactoryDal/userManager/register'
        self.predict_url = self.base_url + '/nerfactory/predict'
        self.parser_get_url = self.base_url + '/nerFactoryDal/parserLibraryList'


    def register_tde_user(self, user_name=None):
        if user_name is None:
            user_name = self.app_id
        params = {
            'userId': self.app_id,
            'userName': user_name
        }
        try:
            response = requests.post(url=self.tde_user_register_url, json=params)
            if self._response_failed(response):
                logging.error("ner: 初始化解析器用户失败: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                response_json = response.json()
                return response_json.get('status') == 'success'
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.tde_user_register_url, traceback.format_exc())
            logging.error('ner: 初始化解析器用户失败， 原因：' + msg)

    def predict(self, sentence: str, parser_id: str):
        params = {
            "companyId": self.app_id,
            "parserId": parser_id,
            "sentence": sentence
        }
        try:
            response = requests.post(url=self.predict_url, json=params)
            if self._response_failed(response):
                logging.error("ner: 调用解析器失败: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                return response.json()['data']['predictResult']
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.predict_url, traceback.format_exc())
            logging.error('ner: 调用解析器失败， 原因：' + msg)

    def get_parser_library(self):
        try:
            response = requests.get(url=self.parser_get_url + "?userId=" + self.app_id)
            if self._response_failed(response):
                logging.error("ner: 调用解析器失败: 接口异常。" + str(response.json()) if response is not None else '')
            else:
                return response.json()['data']
        except Exception:
            msg = "Exception When POST '{}' : {}".format(self.parser_get_url, traceback.format_exc())
            logging.error('ner: 调用解析器失败， 原因：' + msg)



import os
import random
import time
import traceback
import copy

from tqdm import tqdm

from .module import Module
from ..caller.knowledge_caller import KnowledgeGraphCaller
from ..entity.exception import BfEngineException
from ..logger import log

data_template = {
    "data": {
        "entities": [],
        "properties": [],
        "entity_properties": [],
        "synonyms": {"entity": [], "property": [], "value": []},
        "introductions": {"property": [], "entity": {"synonyms": [], "corpus": [], "introduction": []}},
        "values": []
    }
}

entity_template = {
    "id": -1,
    "name": "",
    "sub_entities": [
    ],
    "relation": "has_a"
}
property_template = {
    "id": -1,
    "name": "",
    "corpus": [
    ],
    "type": "文本",
    "unit": "",
    "speech": ""
}
value_template = {
    "id": -1,
    "entity": "",
    "property": "",
    "value": ""
}


class DataGenerator:

    @staticmethod
    def generate_data(raw_data: dict, caller: KnowledgeGraphCaller):
        kg_data = copy.deepcopy(data_template)
        entities = []
        entity_name_set = set()
        property_name_set = set()
        properties = []
        values = []
        kg_data['data']['entities'] = entities
        kg_data['data']['properties'] = properties
        kg_data['data']['values'] = values
        if raw_data and raw_data['data']:
            datas = raw_data['data']
            for data in datas:
                entity_name = data.get('entity')
                if entity_name not in entity_name_set:
                    entity_dict = DataGenerator._get_entity(entity_name)
                    entities.append(entity_dict)
                    entity_name_set.add(entity_name)
            for data in datas:
                property_name = data.get('property')
                value = data.get('value')
                unit = data.get('unit')
                if value in entity_name_set:
                    entity_type = True
                else:
                    value_entity = caller.load_entity(0, value)
                    if value_entity is not None and value_entity.get('type') == 1:
                        entity_type = True
                    else:
                        entity_type = False
                property_dict = DataGenerator._get_property(property_name, entity_type, unit)
                properties.append(property_dict)
                property_name_set.add(property_name)

            for data in datas:
                entity_name = data.get('entity')
                property_name = data.get('property')
                value = data.get('value')
                value_dict = DataGenerator._get_value(entity_name, property_name, value)
                values.append(value_dict)
        return kg_data

    @staticmethod
    def _get_property(property_name, entity_type, unit):
        if entity_type and unit is not None:
            log.error(property_name + '实体类型属性不能有单位')
        property_dict = copy.deepcopy(property_template)
        property_dict['name'] = property_name
        property_dict['corpus'].append("的" + property_name)
        if entity_type:
            property_dict['type'] = '实体'
        elif unit:
            property_dict['unit'] = unit
            property_dict['type'] = '数值'
        elif len(property_name) <= 20:
            property_dict['type'] = '关键字'
        return property_dict

    @staticmethod
    def _get_entity(entity_name):
        entity_dict = copy.deepcopy(entity_template)
        entity_dict['name'] = entity_name
        return entity_dict

    @staticmethod
    def _get_value(entity_name, property_name, value):
        value_dict = copy.deepcopy(value_template)
        value_dict['entity'] = entity_name
        value_dict['property'] = property_name
        value_dict['value'] = value
        return value_dict


class KnowledgeGraph(Module):
    """
    知识图谱
    """

    def __init__(self, app_id, set, init):
        super().__init__(app_id, 'kg', set)
        self.caller = KnowledgeGraphCaller(app_id)
        if init:
            self.init()

    def init(self):
        self.caller.init_data()

    # 下面是 second version 增量训练

    def add_triple_value(self, data: dict) -> bool:
        """
        增量批量添加三元组数值
        :param data: 三元组数据
        "return: 添加是否成功
        """
        try:
            log.info('kg: batch add triple')
            # 整理数据
            data = DataGenerator.generate_data(data, self.caller)
            self._upload_triple_value(data)

        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error("kg: batch add triple error")
            log.error(e)
            traceback.print_exc()
            return False

    def add_intro(self, data: dict) -> bool:
        """
        增量批量添加实体和属性简介
        :param data: 简介数据格式 {type, name, intro}
        "return: 添加是否成功
        """
        try:
            log.info('kg: batch add intro')
            # 整理数据
            for single in data['data']:
                if single['type'] == 'entity':
                    entity = self.caller.load_entity(0, single['name'])
                    if entity is not None:
                        result = self.caller.update_entity_intro(entity.get('id'), single['intro'])
                        if not result:
                            log.error("更新实体简介失败：" + single['name'])
                elif single['type'] == 'property':
                    property = self.caller.load_property(single['name'])
                    if property is not None:
                        result = self.caller.update_property_intro(property.get('id'), single['intro'])
                        if not result:
                            log.error("更新属性简介失败：" + single['name'])
            return True
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error("kg: batch add intro error")
            log.error(e)
            traceback.print_exc()
            return False

    def add_property_corpus(self, data: dict) -> bool:
        """
        增量批量添加属性语料
        :param data: 语料数据格式 {name, corpus}
        "return: 添加是否成功
        """
        try:
            log.info('kg: batch add property corpus')
            # 整理数据
            for single in data['data']:
                property = self.caller.load_property(single['name'])
                if property is not None:
                    result = self.caller.add_property_corpus(property, single['corpus'])
                    if not result:
                        log.error("添加属性语料失败：" + single['name'])
            return True
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error("kg: batch add property corpus error")
            log.error(e)
            traceback.print_exc()
            return False

    def update_property_speech(self, data: dict) -> bool:
        """
        增量批量添加属性语料
        :param data: 语料数据格式 {name, speech}
        "return: 添加是否成功
        """
        try:
            log.info('kg: batch add property speech')
            # 整理数据
            for single in data['data']:
                property = self.caller.load_property(single['name'])
                if property is not None:
                    result = self.caller.update_property_speech(property, single['speech'])
                    if not result:
                        log.error("更新属性话术失败：" + single['name'])
            return True
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error("kg: batch update property speech error")
            log.error(e)
            traceback.print_exc()
            return False

    def add_word_synonym(self, word_name: str, synonym: list):
        return self.caller.add_word_synonym(word_name, synonym)

    def update_word_synonym(self, word_name: str, synonym: list):
        return self.caller.update_word_synonym(word_name, synonym)

    def delete_word_synonym(self, word_name):
        return self.caller.delete_word_synonym(word_name)

    def get_word_synonym(self, word_name):
        return self.caller.get_word_synonym(word_name)

    def publish(self):
        """
        发布KG 从线下到线上
        :return: 训练是否成功
        """
        try:
            self._sync_with_bar()
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error("kg: 发布异常")
            log.error(e)
            traceback.print_exc()
            return False

    # 下面是 first version
    def train(self, data: dict = None, path: str = None) -> bool:
        """
        训练kg  传入data(json) 或者 path(数据文件) 同时传入优先取data

        如果已经导入过数据， data和path可以为空， 表示直接开始训练
        :param path: 导入文件的路径
        :param data: 训练数据
        :return: 训练是否成功
        """
        try:
            log.info('kg: prepare train')
            if data is not None:
                data = DataGenerator.generate_data(data, self.caller)
                self._upload_data(data)
            elif path is not None:
                if os.path.exists(path):
                    self._upload_file(path, True)
                else:
                    log.warning('file not fond in path ' + path)
                    return False
            self._train()
            self.caller.sync_data()
            time.sleep(2)
            if data is not None or path is not None:
                self._sync()
            return True
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error("kg: 训练异常")
            log.error(e)
            traceback.print_exc()
            return False

    def train_by_path(self, data_path: str, use_old: bool = None) -> bool:
        """
        训练kg
        :param data_path: 训练数据文件路径
        :return: 训练是否成功
        """
        try:
            self._upload_file(data_path, use_old)
            self._train()
            self._sync()
            return True
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error("kg: 训练异常")
            log.error(e)
            traceback.print_exc()
            return False

    def _upload_data(self, data):
        # log.info("kg: 开始上传训练数据")
        self.caller.load_data(data)
        # log.info("kg: 训练数据上传成功")

    def _upload_file(self, file_path, use_old):
        # log.info("kg: 开始上传训练数据")
        self.caller.load_file(file_path, use_old)
        # log.info("kg: 训练数据上传成功")

    def _upload_triple_value(self, data):
        # log.info("kg: upload triple value")
        # 更新实体
        for entity in data['data']['entities']:
            result = self.caller.load_entity(0, entity['name'])
            if result is None:
                entity_id = self.caller.add_entity(entity['name'])
                if -1 == entity_id:
                    log.error(self, entity + '实体新增失败')
                    continue
            else:
                entity_id = result.get('id')
            entity['id'] = entity_id
        # 更新属性
        i = 0
        for property in data['data']['properties']:

            if property['type'] == '关键字':
                category = 5
            elif property['type'] == '文本':
                category = 2
            elif property['type'] == '数值':
                category = 3
            elif property['type'] == '实体':
                category = 1

            result = self.caller.load_property(property['name'])
            if result is None:
                property_id = self.caller.add_property(property['name'], category, property['unit'])
                if -1 == property_id:
                    log.error(self, property + '属性新增失败')
                    continue
            else:
                property_id = result.get('id')
                # 更新补充属性：
                if str(category) != result.get('category') or property['unit'] != result.get('unit'):
                    entity = data['data']['entities'][i]
                    extra_property = self.caller.load_extra_property(property['name'], entity['name'])
                    if extra_property is None:
                        result = self.caller.add_extra_property(property_id, entity['id'], category,
                                                                property['unit'])
                        if not result:
                            log.error(
                                'kg: add extra property failed, entity=' + entity['name'] + ', property=' + property[
                                    'name'])
                    else:
                        result = self.caller.update_extra_property(property['name'], entity['name'], category,
                                                                   property['unit'])
                        if not result:
                            log.error(
                                'kg: add extra property failed, entity=' + entity['name'] + ', property=' + property[
                                    'name'])
            property['id'] = property_id
            i+=1
        # 更新值
        for entity_property in data['data']['values']:
            entity = self.caller.load_entity(0, entity_property['entity'])
            property = self.caller.load_property(entity_property['property'])
            # 实体类型怎么处理
            relation = self.caller.load_value(entity.get('id'), property.get('id'))
            if not relation:
                value_id = None
                if property.get('category') == '1':
                    value_entity = self.caller.load_entity(0, entity_property['value'])
                    if value_entity:
                        value_id = value_entity.get('id')
                value = entity_property['value']
                self.caller.add_value(entity, property, value_id, value)
            # else:
            #     self.caller.update_value(relation.get('id'), entity, property, relation.get('idTo'), entity_property['value'])
        # log.info("kg: upload triple value success")

    def _train(self):
        log.info("kg: start training")
        self.caller.trigger_train()
        progress = 0
        completed = False
        with tqdm(total=100) as bar:
            bar.set_description("kg training...")
            while not completed:
                completed = self.caller.status('train')
                if progress < 100:
                    inc = random.randint(1, 4)
                    bar.update(inc)
                    progress += inc
                time.sleep(1)
            bar.update(100 - progress)
            bar.set_description("kg train finished")

    def _sync(self):
        # log.info("kg: 开始同步数据")
        self.caller.sync_sandbox()
        # log.info("kg: 同步成功")
        time.sleep(2)
        completed = False
        while not completed:
            completed = self.caller.status('sync')
            # log.info('kg: 数据同步中..')
            time.sleep(1)
        # log.info("kg: 同步出话层数据")
        self.caller.sync_data()
        # log.info("kg: 同步成功")
        time.sleep(2)

    def _sync_with_bar(self):
        log.info("kg: start publishing")
        # log.info("kg: 开始同步数据")
        self.caller.sync_sandbox()
        # log.info("kg: 同步成功")
        time.sleep(2)
        progress = 0
        completed = False
        with tqdm(total=100) as bar:
            bar.set_description("kg publishing...")
            while not completed:
                completed = self.caller.status('sync')
                if progress < 100:
                    inc = random.randint(1, 4)
                    bar.update(inc)
                    progress += inc
                time.sleep(1)
            self.caller.sync_data()
            time.sleep(2)
            bar.update(100 - progress)
            bar.set_description("kg publish finished")

    # 预测出话
    def query(self, text, online: bool = None):
        """
        :param msg: 用户问
        :online: 线上 True 线下 False
        :return: 机器人回答
        """
        return self.caller.call(text, online)


# Generates request data payload to be sent as 'form-data'

class RequestFormData:
    REQUEST_FORM_DATA_BOUNDARY = "REQUEST_FORM_DATA_BOUNDARY"
    FORM_DATA_STARTING_PAYLOAD = '--{0}\r\nContent-Disposition: form-data; name=\\"'.format(REQUEST_FORM_DATA_BOUNDARY)
    FORM_DATA_MIDDLE_PAYLOAD = '\"\r\n\r\n'
    FORM_DATA_ENDING_PAYLOAD = '--{0}--'.format(REQUEST_FORM_DATA_BOUNDARY)
    REQUEST_CUSTOM_HEADER = {
        'content-type': "multipart/form-data; boundary={}".format(REQUEST_FORM_DATA_BOUNDARY),
        'Content-Type': "",
        'cache-control': "no-cache"
    }
    @classmethod
    def generate_form_data_payload(self,kwargs):
        payload = ''
        for key, value in kwargs.items():
            payload += '{0}{1}{2}{3}\r\n'.format(RequestFormData.FORM_DATA_STARTING_PAYLOAD, key, RequestFormData.FORM_DATA_MIDDLE_PAYLOAD, value)
        payload += RequestFormData.FORM_DATA_ENDING_PAYLOAD
        return payload


class Answer:
    """
    机器人回答
    """
    def __init__(self, text: str,related:list=[],score: float=0.0):
        self.text = text
        self.related = related
        self.score = score

    def __str__(self):
        if self.related:
            return 'text: {},related: {}, score: {}'.format(self.text,self.related, self.score if self.score else 0.0)
        return 'text: {}, score: {}'.format(self.text,self.score if self.score else 0.0)
    def __repr__(self):
        return self.__str__()

    def keys(self):
        return ('text', 'related', 'score')
    def __getitem__(self, item):
        return getattr(self, item)



import traceback

from ..entity.exception import BfEngineException, ErrorCode
from ..logger import log


class SetConfig():
    """
    机器人配置
    """
    def __init__(self):
        self.data = {}
    def add(self, name: str,value):
        """
        :param name: 字典key
        :param value: 字典value
        :return: 训练是否成功
        """
        try:
            if name in self.data.keys():
                self.data.pop(name)
            self.data[name]=value
        except BfEngineException as bfe:
            log.error(bfe)
            return False
        except Exception as e:
            log.error(e)
            traceback.print_exc()
            return False
        return True
    def get(self,key:str=None):
        if key in self.data.keys():
            return self.data[key]
        return None
    def exits(self,key:str=None) -> bool:
        if not self.data:
            return False
        return key in self.data.keys()
    def clear(self):
        self.data.clear()




import os

from .util import random_str


class   Config:
    default_app_id = os.environ.get('BF_ENGINE_DEFAULT_APP_ID', 'emotibotisthebestrobotoftheworld')
    base_url = os.environ.get('BF_ENGINE_BASE_URL', 'https://capricorn.emotibot.com')
    user_id = os.environ.get('BF_ENGINE_DEFAULT_USER_ID', 'bf-engine-user')
    session_id = os.environ.get('BF_ENGINE_DEFAULT_SESSION_ID', 'bf-engine-session' + random_str(8))
    robot_url = os.environ.get('BF_ENGINE_ROBOT_URL', 'openapi/v1/robot')
    chat_url = os.environ.get('BF_ENGINE_CHAT_URL', 'chatbot/#/robotchat')
    remote_url = os.environ.get('BF_ENGINE_REMOTE_URL', 'https://capricorn.emotibot.com')
    backfill_speech = '我还在学习'



from .dialogue_manager import DialogueManager
from .task_engine import TaskEngine
from .knowledge_graph import KnowledgeGraph
from .question_answering import QuestionAnswering
from .intent_answering import IntentAnswering
from .ner_parser import NERParser
from .dialogue_act import DialogueAct
from .config_manager import ConfigManager
from .set import Set


class Bot:
    """
    机器人
    """

    def __init__(self, app_id, init_module: bool):
        self.app_id = app_id
        self.set = Set()
        self.intent = IntentAnswering(app_id, self.set, init_module)
        self.qa = QuestionAnswering(app_id, self.set, init_module)
        self.kg = KnowledgeGraph(app_id, self.set, init_module)
        self.te = TaskEngine(app_id, self.set, init_module)
        self.dm = DialogueManager(app_id, self.set, init_module)
        self.ner = NERParser(app_id, self.set, init_module)
        self.act = DialogueAct(app_id, self.set, init_module)
        self.config = ConfigManager(app_id, self.set, init_module)



import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="bfengine",
    version="0.1.105",
    author="JeremyXin",
    author_email="chengjiexin@emotibot.com",
    description="bf engine sdk for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={
        '': ['data/*']
    },
    install_requires=[
        "requests",
        "tqdm"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    encoding='utf-8'
)



import requests

import json
from .base import CallerBase
from ..config import Config
from ..entity.exception import BfEngineException
from ..utils.request_form_data import RequestFormData


class IntentCaller(CallerBase):
    """
    QA api调用
    """

    def __init__(self, app_id):
        super().__init__(app_id)
        self.app_id = app_id
        self.module = 'qa'
        self.header = {
            "X-locale": "zh-cn",
            "app_id": app_id,
            "X-Appid": app_id,
            "user_id": "bf-engine-sdk",
            "Authorization": "Bearer EMOTIBOTDEBUGGER",
            "Accept": "application/json,text/plain, */*"
        }

        self.upload_intent_json_open_default_group_url = Config.base_url + "/intent/v3/updateIntentGroupSwitch?groupType=1&onOff=true"
        self.upload_intent_json_close_default_group_url = Config.base_url + "/intent/v3/updateIntentGroupSwitch?groupType=1&onOff=false"
        self.upload_intent_json_add_group_url = Config.base_url+"/intent/v3/addIntentGroup"
        self.upload_intent_json_list_group_url = Config.base_url + "/intent/v3/getIntentGroupStatus"

        self.upload_intent_json_list_intents_url = Config.base_url + "/intent/v3/intents"

        self.upload_intent_form_add_intent_url    = Config.base_url+"/api/v3/intents/intent"
        self.upload_intent_json_update_intent_url = Config.base_url+"/intent/v3/update/"

        self.upload_intent_json_train_url = Config.base_url+"/api/v3/intents/train?engine=intent_engine"
        self.upload_intent_json_train_status_url = Config.base_url + "/api/v3/intents/status"
        self.upload_intent_json_predict_intent_url = Config.base_url + "/adm/intent/v3/predict"

        # self.upload_intent_json_open_default_group_url = "http://172.16.102.160" + "/intent/v3/updateIntentGroupSwitch?groupType=1&onOff=true"
        # self.upload_intent_json_add_group_url = "http://172.16.102.160" + "/intent/v3/addIntentGroup"
        # self.upload_intent_json_list_group_url = "http://172.16.102.160" + "/intent/v3/getIntentGroupStatus"
        #
        # self.upload_intent_json_list_intents_url = "http://172.16.102.160" + "/intent/v3/intents"
        #
        # self.upload_intent_form_add_intent_url = "http://172.16.102.160" + "/api/v3/intents/intent"
        # self.upload_intent_json_update_intent_url = "http://172.16.102.160" + "/intent/v3/update/"
        #
        # self.upload_intent_json_train_url = "http://172.16.102.160" + "/api/v3/intents/train?engine=intent_engine"
        # self.upload_intent_json_train_status_url = "http://172.16.102.160" + "/api/v3/intents/status"
        # self.upload_intent_json_predict_intent_url = "http://172.16.102.160" + "/adm/intent/v3/predict"


    def open_default_intent(self):
        group_ids = []
        group_obj_list = self.intent_group_list()
        for group_obj in group_obj_list:
            group_ids.append(group_obj["groupId"])
        for groud_id in group_ids:
            resp = requests.get(self.upload_intent_json_open_default_group_url+"&groupId="+groud_id,headers=self.header, json={}).json()
            code = resp["status"]

    def close_default_intent(self):
        group_ids = []
        group_obj_list = self.intent_group_list()
        for group_obj in group_obj_list:
            group_ids.append(group_obj["groupId"])
        for group_id in group_ids:
            resp = requests.get(self.upload_intent_json_close_default_group_url + "&groupId=" + group_id,
                                headers=self.header, json={}).json()
            code = resp["status"]

    def open_intent_group(self, group_id: str = None):
        resp = requests.get(self.upload_intent_json_open_default_group_url + "&groupId=" + group_id,
                            headers=self.header, json={}).json()
    def close_intent_group(self,group_id:str=None):
        resp = requests.get(self.upload_intent_json_close_default_group_url + "&groupId=" + group_id,
                            headers=self.header, json={}).json()
        code = resp["status"]
    def add_group(self,domain:str):
        """
        添加意图集
        :return 返回添加后的意图group ID集
        """
        group_ids = []
        exist_group_name_list = []  # 已存在意图组名称
        try:
            group_ids.append(self._group_add(domain))
        except BfEngineException as bfe:
            exist_group_name_list.append(domain)
        exist_group_list = self._intent_group_search(exist_group_name_list)
        for exist_group in exist_group_list:
            if "groupId" in exist_group.keys():
                group_ids.append(exist_group["groupId"])
        return group_ids[0] if len(group_ids)>0 else None
    def add_intent(self,group_id:str=None,intent:str=None):
        intent_ids = []
        exist_intent_name_list = []  # 已存在意图组名称
        try:
            self._intent_add(group_id=group_id, intent=intent)
        except BfEngineException as bfe:
            exist_intent_name_list.append(intent["name"])

        exist_intents_list = self._intent_intents_search(group_id=group_id,intents=exist_intent_name_list)
        for exist_intent in exist_intents_list:
            if "id" in exist_intent.keys():
                intent_ids.append(exist_intent["id"])
        return intent_ids[0] if len(intent_ids) > 0 else None
    def intent_group_list(self):
        """
        预测标准问
        :param text: 用户query
        :param online: 线上|线下
        :return 上传id
        """
        resp = requests.get(self.upload_intent_json_list_group_url + "?appId=" + self.app_id+"&&groupType=0", headers=self.header, json={}).json()

        # 意图$组名称
        code = int(resp["status"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["result"]

    def intent_intents_list(self,group_id:str=None):
        """
        预测标准问
        :param text: 用户query
        :param online: 线上|线下
        :return 上传id
        """
        resp = requests.get(self.upload_intent_json_list_intents_url + "?groupId=" + group_id,
                            headers=self.header, json={}).json()

        # 意图$组名称
        code = int(resp["status"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["result"]["intentList"] if "result" in resp.keys() else []
    def _group_add(self, domain):

        data = {
            "groupName": domain,
            "appId": self.app_id
        }
        resp = requests.post(self.upload_intent_json_add_group_url, headers=self.header, json=data).json()

        # 意图$组名称
        code = int(resp["status"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["result"]
    def _intent_group_search(self, groups):
        """
        查询已经存在标签的IDS
        :param 查询tags
        :return 标签列表
        """
        group_objs = []
        group_obj_list = self.intent_group_list()
        for group in groups:
            for group_obj in group_obj_list:
                if group == group_obj["name"]:
                    group_objs.append(group_obj)
        return group_objs
    def _intent_add(self,group_id:str=None,intent:str=None):
        params = {
            "innerIntentId": -1,
            "groupId": group_id,
            "name": intent["name"],
            "positive": json.dumps(intent["corpus"]["positive"]),
            "negative": json.dumps(intent["corpus"]["negative"])
        }
        resp = requests.post(self.upload_intent_form_add_intent_url,headers=self.header, data=params).json()
        code = int(resp["status"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["result"]

    def _intent_intents_search(self, group_id:str=None,intents:list=[]):
        intent_objs = []
        intent_obj_list = self.intent_intents_list(group_id=group_id)
        for intent_name in intents:
            for intent_obj in intent_obj_list:
                if intent_name == intent_obj["name"]:
                    intent_objs.append(intent_obj)
        return intent_objs

    def train(self,group_id:str=None):
        data = {}
        resp = requests.post(self.upload_intent_json_train_url+ "&groupId=" + group_id, headers=self.header, json=data).json()
        # 问题$答案上传进度
        code = int(resp["status"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["result"]
    def train_status(self,group_id:str=None):
        resp = requests.get(self.upload_intent_json_train_status_url+"?groupId=" + group_id).json()

        code = int(resp["status"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        status = resp["result"]["status"]
        if status == "NEED_TRAIN":
            return 100
        return resp["result"]["progress"]
    def predict(self, text: str, online: bool = False) -> list:
        group_ids = []
        group_obj_list = self.intent_group_list()
        for group_obj in group_obj_list:
            group_ids.append(group_obj["groupId"])

        """
        预测标准问
        :param text: 用户query
        :param online: 线上|线下
        :return 上传id
        """
        data = {
            "sentence": text,
            "app_id": self.app_id,
            "group_ids":group_ids
        }
        resp = requests.post(self.upload_intent_json_predict_intent_url, headers=self.header, json=data).json()
        status = resp["status"]
        if status == "OK":
            return resp["predictions"]
        # 问题$答案上传进度
        code = int(resp["status"])
        msg = str(resp["message"])
        if code != 0:
            raise BfEngineException(code=code, msg=msg)
        return resp["result"]
